# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

from struct import pack


class packer():
    def __init__(self, default_size_char='Q') -> None:
        self.size_char = default_size_char

    def custom_pack(self, num_of_chunks: int, expected_value: int, expected_value_index: int = 0) -> bytes:
        input = [0] * num_of_chunks
        input[expected_value_index] = expected_value
        return pack(f'{num_of_chunks}{self.size_char}', *input)

    def pack_pci(self, expected_value: int) -> bytes:
        return self.custom_pack(5, expected_value, 4)

    def pack_cpuinfo(self, expected_value: int) -> bytes:
        return self.custom_pack(4, expected_value, 0)

    def pack_ioport(self, expected_value: int) -> bytes:
        return self.custom_pack(3, expected_value, 2)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Trusted Platform Module (TPM) HAL component

https://trustedcomputinggroup.org
"""

import struct
from typing import Tuple

from chipsec.library.logger import print_buffer
from chipsec.hal import hal_base
from chipsec.library.tpm import tpm_defines


class TPM_BASE(hal_base.HALBase):
    def __init__(self, cs):
        super(TPM_BASE, self).__init__(cs)
        self.TPM_BASE = self.cs.Cfg.MEMORY_RANGES['TPM']['address']
        self.access_address = 0x0

    # def get_registers(self) -> list:
    #    return self.list_of_registers

    def request_locality(self, Locality: int) -> bool:
        self.access_address = self.TPM_BASE | Locality | tpm_defines.TPM_ACCESS
        if self.helper.read_mmio_reg(self.access_address, 4) == tpm_defines.BEENSEIZED:
            self.helper.write_mmio_reg(self.access_address, 4, tpm_defines.REQUESTUSE)
            return True
        return False

    def command(self, commandName: str, locality: str, *command_argv: str) -> None:
        raise NotImplementedError()

    def send_command(self, Locality: int, command: bytes, size: int) -> None:
        raise NotImplementedError()

    def read_response(self, Locality: int) -> Tuple[tpm_defines.tpm1_commands.TPM_RESPONSE_HEADER, bytes, bytearray, bytearray]:
        raise NotImplementedError()


class TPM1(TPM_BASE):
    def __init__(self, cs):
        super(TPM1, self).__init__(cs)
        self.helper = cs.helper

    def command(self, commandName: str, locality: str, *command_argv: str) -> None:
        """Send command to the TPM and receive data"""
        try:
            Locality = tpm_defines.LOCALITY[locality]
        except:
            if self.logger.HAL: self.logger.log_bad("Invalid locality value\n")
            return

        requestedUse = self.request_locality(Locality)

        #
        # Build command (big endian) and send/receive
        #
        (command, size) = tpm_defines.COMMANDS_12[commandName](command_argv)
        self.send_command(Locality, command, size)

        (header, _, _, data_blob) = self.read_response(Locality)
        self.logger.log(str(header))
        print_buffer(str(data_blob))
        self.logger.log('\n')

        #
        # Release locality if needed
        #
        if requestedUse:
            self.helper.write_mmio_reg(self.access_address, 4, tpm_defines.BEENSEIZED)
        self.helper.write_mmio_reg(self.access_address, 1, tpm_defines.ACTIVELOCALITY)

    def send_command(self, Locality: int, command: bytes, size: int) -> None:
        """Send a command to the TPM using the locality specified"""
        count = 0

        datafifo_address = self.TPM_BASE | Locality | tpm_defines.TPM_DATAFIFO
        sts_address = self.TPM_BASE | Locality | tpm_defines.TPM_STS
        self.access_address = self.TPM_BASE | Locality | tpm_defines.TPM_ACCESS

        self.helper.write_mmio_reg(self.access_address, 1, tpm_defines.REQUESTUSE)
        #
        # Set status to command ready
        #
        sts_value = self.helper.read_mmio_reg(sts_address, 1)
        while (0 == (sts_value & tpm_defines.COMMANDREADY)):
            self.helper.write_mmio_reg(sts_address, 1, tpm_defines.COMMANDREADY)
            sts_value = self.helper.read_mmio_reg(sts_address, 1)

        while count < size:
            sts_value = self.helper.read_mmio_reg(sts_address, 4)
            burst_count = ((sts_value>>8) & 0xFFFFFF)
            burst_index = 0
            while (burst_index < burst_count) and (count < size):
                datafifo_value = command[count]
                self.helper.write_mmio_reg(datafifo_address, 1, datafifo_value)
                count += 1
                burst_index += 0x1

        self.helper.write_mmio_reg(sts_address, 1, tpm_defines.TPMGO)

    def read_response(self, Locality: int) -> Tuple[tpm_defines.tpm1_commands.TPM_RESPONSE_HEADER, bytes, bytearray, bytearray]:
        """Read the TPM's response using the specified locality"""
        count = 0
        header = ""
        header_blob = bytearray()
        data = ""
        data_blob = bytearray()
        #
        # Build FIFO address
        #
        datafifo_address = self.TPM_BASE | Locality | tpm_defines.TPM_DATAFIFO
        self.access_address = self.TPM_BASE | Locality| tpm_defines.TPM_ACCESS
        sts_address = self.TPM_BASE | Locality| tpm_defines.TPM_STS

        sts_value = self.helper.read_mmio_reg(sts_address, 1)
        data_avail = bin(sts_value & (1<<4))[2]
        #
        # Read data available
        #
        # watchdog?
        while data_avail == '0':
            sts_value = self.helper.read_mmio_reg(sts_address, 1)
            self.helper.write_mmio_reg(sts_address, 1, tpm_defines.DATAAVAIL)
            data_avail = bin(sts_value & (1<<4))[2]

        while count < tpm_defines.HEADERSIZE:
            sts_value = self.helper.read_mmio_reg(sts_address, 4)
            burst_count = ((sts_value>>8) & 0xFFFFFF)
            burst_index = 0
            while (burst_index < burst_count) and (count < tpm_defines.HEADERSIZE):
                header_blob.append(self.helper.read_mmio_reg(datafifo_address, 1))
                count += 1
                burst_index += 0x1

        header = tpm_defines.COMMANDS_12['header'](*struct.unpack_from(tpm_defines.HEADERFORMAT, header_blob))

        count = 0
        if header.DataSize > 10 and header.ReturnCode == 0:
            length = header.DataSize - tpm_defines.HEADERSIZE
            while count < length:
                sts_value = self.helper.read_mmio_reg(sts_address, 4)
                burst_count = ((sts_value>>8) & 0xFFFFFF)
                burst_index = 0
                while (burst_index < burst_count) and (count < length):
                    data_blob.append(self.helper.read_mmio_reg(datafifo_address, 1))
                    count += 1
                    burst_index += 0x1

        return (header, data, header_blob, data_blob)


class TPM2(TPM_BASE):
    def __init__(self, cs, buffer_type = 'fifo_legacy'):
        super(TPM2, self).__init__(cs)
        self.helper = cs.helper
        self.buffer_type = buffer_type

    def command(self, commandName: str, locality: str, *command_argv: str) -> None:
        """Send command to the TPM and receive data"""
        try:
            Locality = tpm_defines.LOCALITY[locality]
        except Exception as e:
            self.logger.log_bad(f"Invalid locality value\n\t{e}")
            return

        requestedUse = self.request_locality(Locality)

        #
        # Build command (big endian) and send/receive
        #
        try:
            (command, size) = tpm_defines.COMMANDS_20[commandName](command_argv)
        except Exception as e:
            self.logger.log_bad(f'Unable to build command\n\t{e}')
            return

        try:
            self.send_command(Locality, command, size)
        except Exception as e:
            self.logger.log_bad(f'Unable to send command\n\t{e}')
            return
        (header, _, _, data_blob) = self.read_response(Locality)
        self.logger.log(str(header))
        print_buffer(str(data_blob))
        self.logger.log('\n')

        #
        # Release locality if needed
        #
        if requestedUse:
            self.helper.write_mmio_reg(self.access_address, 4, tpm_defines.BEENSEIZED)
        self.helper.write_mmio_reg(self.access_address, 1, tpm_defines.ACTIVELOCALITY)

    def send_command(self, Locality: int, command: bytes, size: int) -> None:
        """Send a command to the TPM using the locality specified"""
        count = 0

        datafifo_address = self.TPM_BASE | Locality | tpm_defines.TPM_DATAFIFO
        sts_address = self.TPM_BASE | Locality | tpm_defines.TPM_STS
        self.access_address = self.TPM_BASE | Locality | tpm_defines.TPM_ACCESS

        self.helper.write_mmio_reg(self.access_address, 1, tpm_defines.REQUESTUSE)
        #
        # Set status to command ready
        #
        sts_value = self.helper.read_mmio_reg(sts_address, 1)
        while (0 == (sts_value & tpm_defines.COMMANDREADY)):
            self.helper.write_mmio_reg(sts_address, 1, tpm_defines.COMMANDREADY)
            sts_value = self.helper.read_mmio_reg(sts_address, 1)

        while count < size:
            sts_value = self.helper.read_mmio_reg(sts_address, 4)
            burst_count = ((sts_value>>8) & 0xFFFFFF)
            burst_index = 0
            while (burst_index < burst_count) and (count < size):
                datafifo_value = command[count]
                self.helper.write_mmio_reg(datafifo_address, 1, datafifo_value)
                count += 1
                burst_index += 0x1

        self.helper.write_mmio_reg(sts_address, 1, tpm_defines.TPMGO)

    def read_response(self, Locality: int) -> Tuple[tpm_defines.tpm2_commands.TPM_RESPONSE_HEADER, bytes, bytearray, bytearray]:
        """Read the TPM's response using the specified locality"""
        count = 0
        header = ""
        header_blob = bytearray()
        data = ""
        data_blob = bytearray()
        #
        # Build FIFO address
        #
        datafifo_address = self.TPM_BASE | Locality | tpm_defines.TPM_DATAFIFO
        self.access_address = self.TPM_BASE | Locality| tpm_defines.TPM_ACCESS
        sts_address = self.TPM_BASE | Locality| tpm_defines.TPM_STS

        sts_value = self.helper.read_mmio_reg(sts_address, 1)
        data_avail = bin(sts_value & (1<<4))[2]
        #
        # Read data available
        #
        # watchdog?
        while data_avail == '0':
            sts_value = self.helper.read_mmio_reg(sts_address, 1)
            self.helper.write_mmio_reg(sts_address, 1, tpm_defines.DATAAVAIL)
            data_avail = bin(sts_value & (1<<4))[2]

        while count < tpm_defines.HEADERSIZE:
            sts_value = self.helper.read_mmio_reg(sts_address, 4)
            burst_count = ((sts_value>>8) & 0xFFFFFF)
            burst_index = 0
            while (burst_index < burst_count) and (count < tpm_defines.HEADERSIZE):
                header_blob.append(self.helper.read_mmio_reg(datafifo_address, 1))
                count += 1
                burst_index += 0x1

        header = tpm_defines.COMMANDS_20['header'](*struct.unpack_from(tpm_defines.HEADERFORMAT, header_blob))

        count = 0
        if header.DataSize > 10 and header.ReturnCode == 0:
            length = header.DataSize - tpm_defines.HEADERSIZE
            while count < length:
                sts_value = self.helper.read_mmio_reg(sts_address, 4)
                burst_count = ((sts_value>>8) & 0xFFFFFF)
                burst_index = 0
                while (burst_index < burst_count) and (count < length):
                    data_blob.append(self.helper.read_mmio_reg(datafifo_address, 1))
                    count += 1
                    burst_index += 0x1

        return (header, data, header_blob, data_blob)


# class TPM2_CRB(TPM_BASE):
#     def __init__(self, cs):
#         super(TPM2_CRB, self).__init__(cs)
#         self.helper = cs.helper
#         self.TPM_BASE = self.cs.Cfg.MEMORY_RANGES['TPM']['address']
#         self.list_of_registers = []

#     def send_command(self, Locality: int, command: bytes, size: int) -> None:
#         raise NotImplementedError()

#     def read_response(self, Locality: int) -> Tuple[TPM_RESPONSE_HEADER, bytes, bytearray, bytearray]:
#         raise NotImplementedError()


# class TPM2_FIFO_LEGACY(TPM1):
#     def __init__(self, cs):
#         super(TPM2_FIFO_LEGACY, self).__init__(cs)
#         self.helper = cs.helper
#         self.TPM_BASE = self.cs.Cfg.MEMORY_RANGES['TPM']['address']
#         self.list_of_registers = ['TPM_ACCESS', 'TPM_STS', 'TPM_DID_VID', 'TPM_RID', 'TPM_INTF_CAPABILITY', 'TPM_INT_ENABLE']



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


import re
import os
import json
import traceback
import chipsec.library.logger
from chipsec.library.file import get_main_dir
from chipsec.library.url import url
from chipsec.library.returncode import ModuleResult, generate_hash_id, get_module_ids_dictionary

_importlib = True
try:
    import importlib
except ImportError:
    _importlib = False

MODPATH_RE = re.compile(r'^\w+(\.\w+)*$')


class Module:
    def __init__(self, name):
        self.logger = chipsec.library.logger.logger()
        self.name = name
        self.module = None
        self.mod_obj = None
        self.module_ids = get_module_ids_dictionary()
        self.url = url()

    def __lt__(self, other):
        return self.name < other.name

    def __le__(self, other):
        return self.name <= other.name

    def __gt__(self, other):
        return self.name > other.name

    def __ge__(self, other):
        return self.name >= other.name

    def get_name(self):
        return self.name

    def do_import(self):
        loaded = False
        if not MODPATH_RE.match(self.get_name()):
            self.logger.log_error(f'Invalid module path: {self.name}')
        else:
            try:
                if _importlib:
                    self.module = importlib.import_module(self.name)
                loaded = True
                if self.logger.DEBUG:
                    self.logger.log_good(f'imported: {self.name}')
            except BaseException as msg:
                self.logger.log_error(f"Exception occurred during import of {self.name}: '{str(msg)}'")
                if self.logger.DEBUG:
                    self.logger.log_bad(traceback.format_exc())
                raise msg
        return loaded

    def update_module_ids_file(self):
        with open(os.path.join(get_main_dir(), 'chipsec', 'library', 'module_ids.json'), 'w') as module_ids_file:
            module_ids_file.write(json.dumps(self.module_ids))

    def get_module_id(self, module_name):
        if module_name in self.module_ids:
            module_id = self.module_ids[module_name]
        else:
            module_id = generate_hash_id(module_name)
            self.module_ids[module_name] = module_id
            self.update_module_ids_file()
        return module_id

    def run(self, module_argv):
        self.get_module_object()

        if module_argv:
            self.logger.log(f'[*] Module arguments ({len(module_argv):d}):')
            self.logger.log(module_argv)
        else:
            module_argv = []

        if isinstance(self.mod_obj, chipsec.module_common.BaseModule):
            self.mod_obj.result.id = self.get_module_id(self.name)
            self.mod_obj.result.url = self.url.get_module_url(self.name)
            if self.mod_obj.is_supported():
                result = self.mod_obj.run(module_argv)
            else:
                self.mod_obj.result.setStatusBit(self.mod_obj.result.status.NOT_APPLICABLE)
                result = self.mod_obj.result.getReturnCode(ModuleResult.NOTAPPLICABLE)
                self.logger.log(f'Skipping module {self.name} since it is not applicable in this environment and/or platform')

        return result

    def get_module_object(self):
        if self.mod_obj is None:
            try:
                if _importlib:
                    pkg = getattr(self.module, "__package__")
                    class_name = getattr(self.module, "__name__")
                    if pkg:
                        class_name = class_name.replace(pkg, '')
                    if class_name.startswith('.'):
                        class_name = class_name.replace('.', '')
                    for iname, iref in self.module.__dict__.items():
                        if isinstance(iref, type):
                            if issubclass(iref, chipsec.module_common.BaseModule):
                                if iname.lower() == class_name.lower():
                                    self.mod_obj = iref()
                if self.mod_obj is None:
                    raise ModuleNotFoundError(self.module)
            except (AttributeError, TypeError, ModuleNotFoundError):
                self.logger.chipsecLogger.exception('Error getting module object')

    def get_location(self):
        myfile = ''
        try:
            if _importlib:
                myfile = getattr(self.module, '__file__')
        except:
            pass
        return myfile

    def get_tags(self):
        module_tags = []
        metadata_tags = []
        try:
            if _importlib:
                module_tags = getattr(self.module, 'TAGS')
                metadata_tags = getattr(self.module, 'METADATA_TAGS')
        except:
            pass

        return module_tags, metadata_tags

    def __str__(self):
        return self.get_name()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Checks for SPI Controller Flash Descriptor Security Override Pin Strap (FDOPSS).
On some systems, this may be routed to a jumper on the motherboard.

Usage:
    ``chipsec_main -m common.spi_fdopss``

Examples:
    >>> chipsec_main.py -m common.spi_fdopss

Registers used:
    - HSFS.FDOPSS

"""

from chipsec.library.exceptions import CSReadError
from chipsec.module_common import BaseModule, BIOS
from chipsec.library.returncode import ModuleResult
from typing import List

TAGS = [BIOS]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'SPI_FDOPSS']


class spi_fdopss(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self) -> bool:
        if not self.cs.register.has_field('HSFS', 'FDOPSS'):
            self.logger.log_important('HSFS.FDOPSS field not defined for platform.  Skipping module.')
            return False
        return True

    def check_fd_security_override_strap(self) -> int:
        hsfs_reg = self.cs.register.read('HSFS')
        self.cs.register.print('HSFS', hsfs_reg)
        fdopss = self.cs.register.get_field('HSFS', hsfs_reg, 'FDOPSS')

        if (fdopss != 0):
            self.logger.log_passed('SPI Flash Descriptor Security Override is disabled')
            return self.result.getReturnCode(ModuleResult.PASSED)
        else:
            self.logger.log_failed('SPI Flash Descriptor Security Override is enabled')
            self.result.setStatusBit(self.result.status.CONFIGURATION)
            return self.result.getReturnCode(ModuleResult.FAILED)

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test('SPI Flash Descriptor Security Override Pin-Strap')
        try:
            self.res = self.check_fd_security_override_strap()
        except CSReadError as err:
            self.logger.log_warning(f'Unable to read register: {err}')
            self.result.setStatusBit(self.result.status.VERIFY)
            self.res = self.result.getReturnCode(ModuleResult.WARNING)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
The idt, gdt and ldt commands print the IDT, GDT and LDT, respectively.

IDT command:

>>> chipsec_util idt [cpu_id]

Examples:

>>> chipsec_util idt 0
>>> chipsec_util idt

GDT command:

>>> chipsec_util gdt [cpu_id]

Examples:

>>> chipsec_util gdt 0
>>> chipsec_util gdt

LDT command:

>>> chipsec_util ldt [cpu_id]

Examples:

>>> chipsec_util ldt 0
>>> chipsec_util ldt
"""

from argparse import ArgumentParser

from chipsec.command import BaseCommand, toLoad

# CPU descriptor tables


class IDTCommand(BaseCommand):
    """
    >>> chipsec_util idt [cpu_id]

    Examples:

    >>> chipsec_util idt 0
    >>> chipsec_util idt
    """

    def requirements(self) -> toLoad:
        return toLoad.Driver

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=IDTCommand.__doc__)
        parser.add_argument('_thread', metavar='thread', type=lambda x: int(x, 0), nargs='?', default=None, help="thread")
        parser.parse_args(self.argv, namespace=self)

    def run(self) -> None:
        num_threads = self.cs.msr.get_cpu_thread_count()
        if self._thread and self._thread < num_threads:
            self.logger.log(f'[CHIPSEC] Dumping IDT of CPU thread {self._thread:d}')
            self.cs.msr.IDT(self._thread, 4)
        else:
            self.logger.log(f'[CHIPSEC] Dumping IDT of {num_threads:d} CPU threads')
            self.cs.msr.IDT_all(4)


class GDTCommand(BaseCommand):
    """
    >>> chipsec_util gdt [cpu_id]

    Examples:

    >>> chipsec_util gdt 0
    >>> chipsec_util gdt
    """

    def requirements(self) -> toLoad:
        return toLoad.Driver

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=GDTCommand.__doc__)
        parser.add_argument('_thread', metavar='thread', type=lambda x: int(x, 0), nargs='?', default=None, help="thread")
        parser.parse_args(self.argv, namespace=self)

    def run(self) -> None:
        num_threads = self.cs.msr.get_cpu_thread_count()
        if self._thread and self._thread < num_threads:
            self.logger.log(f'[CHIPSEC] Dumping IDT of CPU thread {self._thread:d}')
            self.cs.msr.GDT(self._thread, 4)
        else:
            self.logger.log(f'[CHIPSEC] Dumping IDT of {num_threads:d} CPU threads')
            self.cs.msr.GDT_all(4)


class LDTCommand(BaseCommand):
    """
    >>> chipsec_util ldt [cpu_id]

    Examples:

    >>> chipsec_util ldt 0
    >>> chipsec_util ldt
    """

    def requirements(self) -> toLoad:
        return toLoad.Nil

    def parse_arguments(self) -> None:
        return

    def run(self) -> None:
        self.logger.log_error("[CHIPSEC] ldt not implemented")


commands = {'idt': IDTCommand, 'gdt': GDTCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Command-line utility providing access to IOMMU engines

>>> chipsec_util iommu list
>>> chipsec_util iommu config [iommu_engine]
>>> chipsec_util iommu status [iommu_engine]
>>> chipsec_util iommu enable|disable <iommu_engine>
>>> chipsec_util iommu pt

Examples:

>>> chipsec_util iommu list
>>> chipsec_util iommu config VTD
>>> chipsec_util iommu status GFXVTD
>>> chipsec_util iommu enable VTD
>>> chipsec_util iommu pt
"""

from chipsec.command import BaseCommand, toLoad
from chipsec.hal import acpi, iommu
from argparse import ArgumentParser
from chipsec.library.exceptions import IOMMUError, AcpiRuntimeError


# I/O Memory Management Unit (IOMMU), e.g. Intel VT-d
class IOMMUCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util iommu', usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_list = subparsers.add_parser('list')
        parser_list.set_defaults(func=self.iommu_list)

        parser_config = subparsers.add_parser('config')
        parser_config.add_argument('engine', type=str, default='', nargs='?', help='IOMMU Engine')
        parser_config.set_defaults(func=self.iommu_config)

        parser_status = subparsers.add_parser('status')
        parser_status.add_argument('engine', type=str, default='', nargs='?', help='IOMMU Engine')
        parser_status.set_defaults(func=self.iommu_status)

        parser_enable = subparsers.add_parser('enable')
        parser_enable.add_argument('engine', type=str, help='IOMMU Engine')
        parser_enable.set_defaults(func=self.iommu_enable)

        parser_disable = subparsers.add_parser('disable')
        parser_disable.add_argument('engine', type=str, help='IOMMU Engine')
        parser_disable.set_defaults(func=self.iommu_disable)

        parser_pt = subparsers.add_parser('pt')
        parser_pt.add_argument('engine', type=str, default='', nargs='?', help='IOMMU Engine')
        parser_pt.set_defaults(func=self.iommu_pt)

        parser.parse_args(self.argv, namespace=self)

    def iommu_list(self) -> None:
        self.logger.log("[CHIPSEC] Enumerating supported IOMMU engine names:")
        self.logger.log(f'{list(iommu.IOMMU_ENGINES.keys())}')
        self.logger.log_important('\nNote: These are the IOMMU engine names supported by iommu_cmd.')
        self.logger.log_important('It does not mean they are supported/enabled in the current platform.')

    def iommu_engine(self, cmd) -> None:
        try:
            _iommu = iommu.IOMMU(self.cs)
        except IOMMUError as msg:
            self.logger.log(msg)
            return

        if self.engine:
            if self.engine in iommu.IOMMU_ENGINES.keys():
                _iommu_engines = [self.engine]
            else:
                self.logger.log_error(f'IOMMU name \'{self.engine}\' not recognized. Run \'iommu list\' command for supported IOMMU names')
                return
        else:
            _iommu_engines = iommu.IOMMU_ENGINES.keys()

        if 'config' == cmd:
            try:
                _acpi = acpi.ACPI(self.cs)
            except AcpiRuntimeError as msg:
                self.logger.log(msg)
                return

            if _acpi.is_ACPI_table_present(acpi.ACPI_TABLE_SIG_DMAR):
                self.logger.log("[CHIPSEC] Dumping contents of DMAR ACPI table..\n")
                _acpi.dump_ACPI_table(acpi.ACPI_TABLE_SIG_DMAR)
            else:
                self.logger.log("[CHIPSEC] Couldn't find DMAR ACPI table\n")

        for e in _iommu_engines:
            if (cmd == 'config'):
                _iommu.dump_IOMMU_configuration(e)
            elif (cmd == 'pt'):
                _iommu.dump_IOMMU_page_tables(e)
            elif (cmd == 'status'):
                _iommu.dump_IOMMU_status(e)
            elif (cmd == 'enable'):
                _iommu.set_IOMMU_Translation(e, 1)
            elif (cmd == 'disable'):
                _iommu.set_IOMMU_Translation(e, 0)

    def iommu_config(self) -> None:
        self.iommu_engine('config')

    def iommu_status(self) -> None:
        self.iommu_engine('status')

    def iommu_enable(self) -> None:
        self.iommu_engine('enable')

    def iommu_disable(self) -> None:
        self.iommu_engine('disable')

    def iommu_pt(self) -> None:
        self.iommu_engine('pt')

    def run(self) -> None:
        self.func()


commands = {'iommu': IOMMUCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Native Linux helper
"""

import mmap
import os
import platform
import resource
import struct
import sys
from typing import Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from ctypes import Array

from chipsec.library import defines
from chipsec.library.exceptions import OsHelperError
from chipsec.helper.basehelper import Helper
from chipsec.helper.linuxnative.cpuid import CPUID
from chipsec.helper.linuxnative.legacy_pci import LegacyPci
from chipsec.library.logger import logger


class MemoryMapping(mmap.mmap):
    """Memory mapping based on Python's mmap.
    This subclass keeps tracks of the start and end of the mapping.
    """

    def __init__(self, fileno, length, flags, prot, offset):
        self.start = offset
        self.end = offset + length
        super().__init__()


class LinuxNativeHelper(Helper):

    DEV_MEM = '/dev/mem'
    DEV_PORT = '/dev/port'

    def __init__(self):
        super(LinuxNativeHelper, self).__init__()
        self.os_system = platform.system()
        self.os_release = platform.release()
        self.os_version = platform.version()
        self.os_machine = platform.machine()
        self.os_uname = platform.uname()
        self.name = 'LinuxNativeHelper'
        self.dev_fh = None
        self.dev_mem = None
        self.dev_port = None
        self.dev_msr = None

        # A list of all the mappings allocated via map_io_space. When using
        # read/write MMIO, if the region is already mapped in the process's
        # memory, simply read/write from there.
        self.mappings = []

###############################################################################################
# Driver/service management functions
###############################################################################################
    def create(self) -> bool:
        logger().log_debug('[helper] Linux Helper created')
        return True

    def start(self) -> bool:
        self.init()
        logger().log_debug('[helper] Linux Helper started/loaded')
        return True

    def stop(self) -> bool:
        self.close()
        logger().log_debug('[helper] Linux Helper stopped/unloaded')
        return True

    def delete(self) -> bool:
        logger().log_debug('[helper] Linux Helper deleted')
        return True

    def init(self):
        x64 = True if sys.maxsize > 2**32 else False
        self._pack = 'Q' if x64 else 'I'

    def devmem_available(self) -> bool:
        """Check if /dev/mem is usable.
           In case the driver is not loaded, we might be able to perform the
           requested operation via /dev/mem. Returns True if /dev/mem is
           accessible.
        """
        if self.dev_mem:
            return True

        try:
            self.dev_mem = os.open(self.DEV_MEM, os.O_RDWR)
            return True
        except IOError as err:
            raise OsHelperError('Unable to open /dev/mem.\n'
                                'This command requires access to /dev/mem.\n'
                                'Are you running this command as root?\n'
                                f'{str(err)}', err.errno)

    def devport_available(self) -> bool:
        """Check if /dev/port is usable.
           In case the driver is not loaded, we might be able to perform the
           requested operation via /dev/port. Returns True if /dev/port is
           accessible.
        """
        if self.dev_port:
            return True

        try:
            self.dev_port = os.open(self.DEV_PORT, os.O_RDWR)
            return True
        except IOError as err:
            raise OsHelperError('Unable to open /dev/port.\n'
                                'This command requires access to /dev/port.\n'
                                'Are you running this command as root?\n'
                                f'{str(err)}', err.errno)

    def devmsr_available(self) -> bool:
        """Check if /dev/cpu/CPUNUM/msr is usable.
           In case the driver is not loaded, we might be able to perform the
           requested operation via /dev/cpu/CPUNUM/msr. This requires loading
           the (more standard) msr driver. Returns True if /dev/cpu/CPUNUM/msr
           is accessible.
        """
        if self.dev_msr:
            return True

        try:
            self.dev_msr = {}
            if not os.path.exists('/dev/cpu/0/msr'):
                os.system('modprobe msr')
            for cpu in os.listdir('/dev/cpu'):
                logger().log_debug(f'found cpu = {str(cpu)}')
                if cpu.isdigit():
                    cpu = int(cpu)
                    self.dev_msr[cpu] = os.open(f'/dev/cpu/{str(cpu)}/msr', os.O_RDWR)
                    logger().log_debug(f'Added dev_msr {str(cpu)}')
            return True
        except IOError as err:
            raise OsHelperError('Unable to open /dev/cpu/CPUNUM/msr.\n'
                                'This command requires access to /dev/cpu/CPUNUM/msr.\n'
                                'Are you running this command as root?\n'
                                'Do you have the msr kernel module installed?\n'
                                f'{str(err)}', err.errno)

    def close(self):
        if self.dev_mem:
            os.close(self.dev_mem)
        self.dev_mem = None

###############################################################################################
# Actual API functions to access HW resources
###############################################################################################

    def read_pci_reg(self, bus: int, device: int, function: int, offset: int, size: int, domain: int = 0) -> int:
        device_name = f'{domain:04x}:{bus:02x}:{device:02x}.{function}'
        device_path = f'/sys/bus/pci/devices/{device_name}/config'
        if not os.path.exists(device_path):
            if offset < 256:
                value = LegacyPci.read_pci_config(bus, device, function, offset)
                if size == 1:
                    value = value & 0xFF
                elif size == 2:
                    value = value & 0xFFFF
                elif size == 4:
                    value = value & 0xFFFF_FFFF
                elif size == 8:
                    value = value & 0xFFFF_FFFF_FFFF_FFFF
                return value
            else:
                raise ValueError('Offset out of bounds')
        try:
            with open(device_path, 'rb') as config:
                config.seek(offset)
                reg = config.read(size)
                reg = defines.unpack1(reg, size)
                return reg
        except IOError as err:
            raise OsHelperError(f'Unable to open {device_path}', err.errno)

    def write_pci_reg(self, bus: int, device: int, function: int, offset: int, value: int, size: int = 4, domain: int = 0) -> int:
        device_name = f'{domain:04x}:{bus:02x}:{device:02x}.{function}'
        device_path = f'/sys/bus/pci/devices/{device_name}/config'
        if not os.path.exists(device_path):
            if offset < 256:
                LegacyPci.write_pci_config(bus, device, function, offset, value)
                return -1
        try:
            with open(device_path, 'wb') as config:
                config.seek(offset)
                config.write(defines.pack1(value, size))
        except IOError as err:
            raise OsHelperError(f'Unable to open {device_path}', err.errno)

        return 0

    # @TODO fix memory mapping and bar_size
    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        if self.devmem_available():
            region = self.memory_mapping(phys_address, size)
            if not region:
                self.map_io_space(phys_address, size, 0)
                region = self.memory_mapping(phys_address, size)
                if not region:
                    logger().log_error(f'Unable to map region {phys_address:08x}')

            # Create memoryview into mmap'ed region
            region_mv = memoryview(region)
            offset_in_region = phys_address - region.start
            if size == 1:
                return region_mv[offset_in_region]

            if offset_in_region % size == 0:
                # Read aligned value
                region_casted = region_mv.cast(defines.SIZE2FORMAT[size])
                return region_casted[offset_in_region // size]

            # Read unaligned value
            return defines.unpack1(region_mv[offset_in_region:offset_in_region + size], size)
        return 0

    # @TODO fix memory mapping and bar_size
    def write_mmio_reg(self, phys_address: int, size: int, value: int) -> None:
        if self.devmem_available():
            reg = defines.pack1(value, size)
            region = self.memory_mapping(phys_address, size)
            if not region:
                self.map_io_space(phys_address, size, 0)
                region = self.memory_mapping(phys_address, size)
                if not region:
                    logger().log_error(f'Unable to map region {phys_address:08x}')

            # Create memoryview into mmap'ed region
            region_mv = memoryview(region)
            offset_in_region = phys_address - region.start
            if size == 1:
                region_mv[offset_in_region] = value
                return

            if offset_in_region % size == 0:
                # Write aligned value
                region_casted = region_mv.cast(defines.SIZE2FORMAT[size])
                region_casted[offset_in_region // size] = value
                return

            # Write unaligned value
            region_mv[offset_in_region:offset_in_region + size] = reg

    def memory_mapping(self, base: int, size: int) -> Optional[MemoryMapping]:
        """Returns the mmap region that fully encompasses this area.
        Returns None if no region matches.
        """
        for region in self.mappings:
            if region.start <= base and region.end >= base + size:
                return region
        return None

    def map_io_space(self, base: int, size: int, cache_type: int) -> None:
        """Map to memory a specific region."""
        if self.devmem_available() and not self.memory_mapping(base, size):
            logger().log_debug(f'[helper] Mapping 0x{base:x} to memory')
            length = max(size, resource.getpagesize())
            page_aligned_base = base - (base % resource.getpagesize())
            mapping = MemoryMapping(self.dev_mem, length, mmap.MAP_SHARED,
                                    mmap.PROT_READ | mmap.PROT_WRITE,
                                    offset=page_aligned_base)
            self.mappings.append(mapping)

    def read_phys_mem(self, phys_address, length: int) -> bytes:
        if self.devmem_available():
            os.lseek(self.dev_mem, phys_address, os.SEEK_SET)
            return os.read(self.dev_mem, length)
        return b'\x00'

    def write_phys_mem(self, phys_address, length: int, newval: bytes) -> int:
        if newval is None:
            return None
        if self.devmem_available():
            os.lseek(self.dev_mem, phys_address, os.SEEK_SET)
            written = os.write(self.dev_mem, newval)
            if written != length:
                logger().log_debug(f'Cannot write {newval} to memory {phys_address:016X} (wrote {written:d} of {length:d})')
            return written
        return -1

    def alloc_phys_mem(self, length, max_phys_address):
        raise NotImplementedError()

    def free_phys_mem(self, physical_address):
        raise NotImplementedError()

    def va2pa(self, va):
        raise NotImplementedError()

    def read_io_port(self, io_port: int, size: int) -> int:
        if self.devport_available():
            os.lseek(self.dev_port, io_port, os.SEEK_SET)

            value = os.read(self.dev_port, size)
            if 1 == size:
                return struct.unpack('B', value)[0]
            elif 2 == size:
                return struct.unpack('H', value)[0]
            elif 4 == size:
                return struct.unpack('I', value)[0]
            else:
                raise ValueError('Invalid size')
        return -1

    def write_io_port(self, io_port: int, value: int, size: int) -> bool:
        if self.devport_available():
            os.lseek(self.dev_port, io_port, os.SEEK_SET)
            if 1 == size:
                fmt = 'B'
            elif 2 == size:
                fmt = 'H'
            elif 4 == size:
                fmt = 'I'
            else:
                raise ValueError('Invalid size')
            written = os.write(self.dev_port, struct.pack(fmt, value))
            if written != size:
                logger().log_debug(f'Cannot write {value} to port {io_port:x} (wrote {written:d} of {size:d})')
                return False
            return True
        return False

    def read_cr(self, cpu_thread_id, cr_number):
        raise NotImplementedError()

    def write_cr(self, cpu_thread_id, cr_number, value):
        raise NotImplementedError()

    def read_msr(self, thread_id: int, msr_addr: int) -> Tuple[int, int]:
        if self.devmsr_available():
            os.lseek(self.dev_msr[thread_id], msr_addr, os.SEEK_SET)
            buf = os.read(self.dev_msr[thread_id], 8)
            unbuf = struct.unpack('2I', buf)
            return (unbuf[0], unbuf[1])
        return (-1, -1)

    def write_msr(self, thread_id: int, msr_addr: int, eax: int, edx: int) -> int:
        if self.devmsr_available():
            os.lseek(self.dev_msr[thread_id], msr_addr, os.SEEK_SET)
            buf = struct.pack('2I', eax, edx)
            written = os.write(self.dev_msr[thread_id], buf)
            if written != 8:
                logger().log_debug(f'Cannot write {buf.hex()} to MSR {msr_addr:x}')
            return written
        return False

    def load_ucode_update(self, cpu_thread_id, ucode_update_buf):
        raise NotImplementedError()

    def get_descriptor_table(self, cpu_thread_id, desc_table_code):
        raise NotImplementedError()

    def EFI_supported(self):
        raise NotImplementedError()

    def get_EFI_variable(self, name, guid):
        raise NotImplementedError()

    def set_EFI_variable(self, name, guid, buffer, buffer_size=None, attrs=None):
        raise NotImplementedError()

    def delete_EFI_variable(self, name, guid):
        raise NotImplementedError()

    def list_EFI_variables(self):
        raise NotImplementedError()

    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        raise NotImplementedError()

    def enum_ACPI_tables(self) -> Optional['Array']:
        raise NotImplementedError()

    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        _cpuid = CPUID()
        return _cpuid(eax, ecx)

    def msgbus_send_read_message(self, mcr, mcrx):
        raise NotImplementedError()

    def msgbus_send_write_message(self, mcr, mcrx, mdr):
        raise NotImplementedError()

    def msgbus_send_message(self, mcr, mcrx, mdr):
        raise NotImplementedError()

    #
    # Affinity functions
    #
    def get_affinity(self) -> Optional[int]:
        try:
            affinity = os.sched_getaffinity(0)
            return list(affinity)[0]
        except Exception:
            return None

    def set_affinity(self, thread_id: int) -> Optional[int]:
        try:
            os.sched_setaffinity(os.getpid(), {thread_id})
            return thread_id
        except Exception:
            return None

    #
    # Logical CPU count
    #
    def get_threads_count(self) -> int:
        import multiprocessing
        return multiprocessing.cpu_count()

    #
    # Send SW SMI
    #
    def send_sw_smi(self, cpu_thread_id, SMI_code_data, _rax, _rbx, _rcx, _rdx, _rsi, _rdi):
        raise NotImplementedError()

    #
    # Hypercall
    #
    def hypercall(self, rcx=0, rdx=0, r8=0, r9=0, r10=0, r11=0, rax=0, rbx=0, rdi=0, rsi=0, xmm_buffer=0):
        raise NotImplementedError()

    #
    # Speculation control
    #
    def retpoline_enabled(self):
        raise NotImplementedError('retpoline_enabled')

    def get_bios_version(self) -> str:
        try:
            filename = '/sys/class/dmi/id/bios_version'
            with open(filename, 'r') as outfile:
                return outfile.read().strip()
        except FileNotFoundError:
            return 'Unable to read bios version'


def get_helper():
    return LinuxNativeHelper()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Common include file for modules

"""

import os
import json
from enum import Enum
from chipsec.library.defines import bit, is_set
from chipsec.library.file import get_main_dir
from chipsec.library.logger import logger
from hashlib import sha256
from typing import Dict


class ReturnCode:
    class status(Enum):
        SUCCESS = [0x0000000000000000, 'Test module completed successfully']
        LOCKS = [bit(31), 'Locks are not set']
        MITIGATION = [bit(30), 'Does not support mitigation']
        CONFIGURATION = [bit(29), 'Configuration not valid']
        PROTECTION = [bit(28), 'Protection not supported/enabled']
        ACCESS_RW = [bit(27), 'Read or write access issues']
        RESTORE = [bit(26), 'Cannot restore binary/value']
        POTENTIALLY_VULNERABLE = [bit(25), 'Found potential vulnerabilities']
        MISMATCH = [bit(24), 'Data does not match expected value']
        REGISTER_NOT_DEFINED = [bit(20), 'Register not defined in Configuration']
        ALL_FFS = [bit(19), 'Read returned all 0xFFs']
        ALL_00S = [bit(18), 'Read returned all 0x00s']
        DEVICE_DISABLED = [bit(17), 'Device is disabled']
        FEATURE_DISABLED = [bit(16), 'Feature is disabled']
        PARSE_ERROR = [bit(15), 'Issue parsing the data']
        UNDEFINED_RANGES = [bit(14), 'Memory ranges are not defined']
        TIMEOUT = [bit(13), 'Operation timed out']
        VERIFY = [bit(12), 'Manual verification/further testing recommended']
        UNSUPPORTED_FEATURE = [bit(11), 'Feature not supported']
        UNSUPPORTED_OPTION = [bit(10), 'Option not supported']
        DEBUG_FEATURE = [bit(9), 'A debug feature is enabled or an unexpected debug state was discovered on this platform']
        NOT_APPLICABLE = [bit(8), 'Skipping module since it is not supported']
        ARCHIVED = [bit(1), 'Archived module, to run use tag ARCHIVED']
        INFORMATION = [bit(0), 'For your information']
        INVALID = [0xFFFFFFFFFFFFFFFF, 'Error running the test']

    def __init__(self, cs):
        self.id = 0x0
        self.url = ''
        self._result = 0x00000000
        self._return_code = self.status.SUCCESS.value[0]
        self._message = ''
        self.logger = logger()
        self.cs = cs

    def setStatusBit(self, status) -> None:
        if is_set(self._result, status.value[0]) is not True:
            self._result ^= status.value[0]
            self._message += ' / ' + status.value[1]

    def setResultBits(self) -> None:
        self._return_code ^= self._result << 32

    def setTestID(self) -> None:
        self._return_code ^= self.id << 4

    def printLogOutput(self) -> None:
        if self._result == self.status.SUCCESS.value[0]:
            self.logger.log_good(f'RC 0x{self._return_code:016x}: {self.status.SUCCESS.value[1]}')
        else:
            self.logger.log_important(f"For next steps: {self.url}")
            self.logger.log_important(f"RC 0x{self._return_code:016x}: {self._message}")

    def buildReturnCode(self, print_output: bool) -> None:
        self.setResultBits()
        self.setTestID()
        if print_output: self.printLogOutput()

    def resetReturnCodeValues(self) -> None:
        self.id = 0x0
        self._result = 0x00000000
        self._return_code = self.status.SUCCESS.value[0]

    def getReturnCode(self, result: int, print_output: bool = True) -> int:
        if self.cs.using_return_codes:
            self.buildReturnCode(print_output)
        else:
            self._return_code = result
        ret_value = self._return_code
        self.resetReturnCodeValues()
        return ret_value


def get_module_ids_dictionary() -> Dict[str, str]:
    with open(os.path.join(get_main_dir(), 'chipsec', 'library', 'module_ids.json'), 'r') as module_ids_file:
        module_ids = json.loads(module_ids_file.read())
    return module_ids


def generate_hash_id(className: str) -> int:
    generated_id = sha256(className.encode("utf-8")).hexdigest()[:7]
    return int(generated_id, 16)

# -------------------------------------------------------
# Legacy results
# -------------------------------------------------------


class ModuleResult(Enum):
    FAILED = 0
    PASSED = 1
    WARNING = 2
    DEPRECATED = 4
    INFORMATION = 5
    NOTAPPLICABLE = 6
    ARCHIVED = 7
    ERROR = -1


result_priority = {
    ModuleResult.PASSED: 0,
    ModuleResult.NOTAPPLICABLE: 0,
    ModuleResult.DEPRECATED: 0,
    ModuleResult.ARCHIVED: 0,
    ModuleResult.INFORMATION: 1,
    ModuleResult.WARNING: 2,
    ModuleResult.FAILED: 3,
    ModuleResult.ERROR: 4
}

ModuleResultName = {
    ModuleResult.FAILED: 'Failed',
    ModuleResult.PASSED: 'Passed',
    ModuleResult.WARNING: 'Warning',
    ModuleResult.DEPRECATED: 'Deprecated',
    ModuleResult.INFORMATION: 'Information',
    ModuleResult.ERROR: 'Error',
    ModuleResult.NOTAPPLICABLE: 'NotApplicable',
    ModuleResult.ARCHIVED: 'Archived',
}


def max_result_priority(previous_result: ModuleResult, current_result: ModuleResult) -> ModuleResult:
    '''Accepts two results and returns either current_result (if equal) or the max of the two'''
    return previous_result if result_priority[previous_result] > result_priority[current_result] else current_result


def getModuleResultName(res, using_return_codes) -> str:
    if using_return_codes:
        result_mask = 0xFFFFFFFF00000000
        status = [ReturnCode.status.SUCCESS.value[0], ReturnCode.status.INFORMATION.value[0], ReturnCode.status.NOT_APPLICABLE.value[0]]
        if ((res & result_mask) >> 32) in status:
            resultName = 'Passed'
        elif ((res & result_mask) >> 32) == ReturnCode.status.ARCHIVED.value[0]:
            resultName = 'Archived'
        else:
            resultName = 'Failed'
        return resultName
    return ModuleResultName[res] if res in ModuleResultName else ModuleResultName[ModuleResult.ERROR]
# -------------------------------------------------------



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Checks protection of UEFI variables defined in the UEFI spec to have certain permissions.

Returns failure if variable attributes are not as defined in `table 11 "Global Variables" <http://uefi.org/>`_ of the UEFI spec.

usage:
    ``chipsec_main -m common.uefi.access_uefispec [-a modify[,attribute]]``

    - ``[-a modify[,attribute]]``: Attempt to modify each variable in addition to checking attributes; optionally change the attribute as well

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -m common.uefi.access_uefispec
    >>> chipsec_main.py -m common.uefi.access_uefispec -a modify
    >>> chipsec_main.py -m common.uefi.access_uefispec -a modify,7

NOTE:
Requires an OS with UEFI Runtime API support.
"""

from chipsec.module_common import BaseModule, SECUREBOOT, BIOS, OPT_MODIFY
from chipsec.library.returncode import ModuleResult
from chipsec.hal.uefi import UEFI, EFI_VARIABLE_NON_VOLATILE, EFI_VARIABLE_BOOTSERVICE_ACCESS, EFI_VARIABLE_RUNTIME_ACCESS, get_attr_string
from chipsec.hal.uefi import EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS, EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS, EFI_VARIABLE_APPEND_WRITE
from chipsec.hal.uefi_common import StatusCode
from typing import List, Optional


TAGS = [BIOS, SECUREBOOT]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'UEFI', 'ACCESS_UEFISPEC']


class access_uefispec(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)
        self._uefi = UEFI(self.cs)

        nv = EFI_VARIABLE_NON_VOLATILE
        bs = EFI_VARIABLE_BOOTSERVICE_ACCESS
        rt = EFI_VARIABLE_RUNTIME_ACCESS
        ta = EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS

        self.uefispec_vars = {
            # From UEFI Spec Table 11 "Global Variables"
            # Updated to version 2.10 Section 3.3 "Globally Defined Variables"
            # https://uefi.org/sites/default/files/resources/UEFI_Spec_2_10_Aug29.pdf
            # https://uefi.org/specs/UEFI/2.10/03_Boot_Manager.html#globally-defined-variables
            "LangCodes": bs | rt,
            "Lang": nv | bs | rt,
            "Timeout": nv | bs | rt,
            "PlatformLangCodes": bs | rt,
            "PlatformLang": nv | bs | rt,
            "ConIn": nv | bs | rt,
            "ConOut": nv | bs | rt,
            "ErrOut": nv | bs | rt,
            "ConInDev": bs | rt,
            "ConOutDev": bs | rt,
            "ErrOutDev": bs | rt,
            "Boot0001": nv | bs | rt,
            "Boot0002": nv | bs | rt,
            "BootOrder": nv | bs | rt,

            "AuditMode": bs | rt,
            "BootNext": nv | bs | rt,
            "BootCurrent": bs | rt,
            "BootOptionSupport": bs | rt,
            "CryptoIndications": nv | bs | rt,
            "CryptoIndicationsSupport": bs | rt,
            "CrytopIndicationsActive": bs | rt,
            "DeployedMode": bs | rt,
            "devAuthBoot": bs | rt,
            "devdbDefault": bs | rt,
            "Driver0001": nv | bs | rt,
            "DriverOrder": nv | bs | rt,
            "Key0001": nv | bs | rt,
            "HwErrRecSupport": nv | bs | rt,  # HwErrRecSupport should be RO
            "SetupMode": bs | rt,  # SetupMode should be RO
            "KEK": nv | bs | rt | ta,
            "OsRecoveryOrder": nv | bs | rt | ta,
            "PK": nv | bs | rt | ta,
            "SignatureSupport": bs | rt,  # RO
            "SecureBoot": bs | rt,  # RO
            "KEKDefault": bs | rt,  # RO
            "PKDefault": bs | rt,  # RO
            "dbDefault": bs | rt,  # RO
            "dbxDefault": bs | rt,  # RO
            "dbtDefault": bs | rt,  # RO
            "OsIndicationsSupported": bs | rt,  # RO only after ExitBootServices()
            "OsIndications": nv | bs | rt,
            "SysPrep0001": nv | bs | rt,
            "SysPrep0002": nv | bs | rt,
            "SysPrepOrder": nv | bs | rt,
            "VendorKeys": bs | rt  # RO
        }

        self.uefispec_ro_vars = ("HwErrRecSupport", "SetupMode", "SignatureSupport", "SecureBoot", "KEKDefault", "PKDefault", "dbDefault", "dbxDefault", "dbtDefault", "VendorKeys")
        uefispec_rt_ro_vars = ("OsIndicationsSupported",)

        if not self.cs.os_helper.is_efi():
            self.uefispec_ro_vars += uefispec_rt_ro_vars

    def is_supported(self) -> bool:
        supported = self.cs.helper.EFI_supported()
        if not supported:
            self.logger.log("OS does not support UEFI Runtime API")
        return supported

    def diff_var(self, data1: int, data2: int) -> bool:
        if data1 is None or data2 is None:
            return data1 != data2

        oldstr = ":".join(f"{c:02x}" for c in data1)
        newstr = ":".join(f"{c:02x}" for c in data2)

        if oldstr != newstr:
            self.logger.log(oldstr)
            self.logger.log(newstr)
            return True
        else:
            return False

    def can_modify(self, name: str, guid: str, data: bytes, attrs: int) -> bool:
        ret = False

        # origdata = _uefi.get_EFI_variable(name, guid)
        origdata = data
        datalen = len(bytearray(data))
        baddata = 'Z' * datalen  # 0x5A is ASCII 'Z'
        if baddata == origdata:
            baddata = 'A' * datalen  # in case we failed to restore previously
        status = self._uefi.set_EFI_variable(name, guid, baddata, len(baddata), attrs)
        if status != StatusCode.EFI_SUCCESS:
            self.logger.log_good(f'Writing EFI variable {name} did not succeed.')
        newdata = self._uefi.get_EFI_variable(name, guid)
        if self.diff_var(newdata, origdata):
            self.logger.log_bad(f'Corruption of EFI variable of concern {name}. Trying to recover.')
            ret = True
            self._uefi.set_EFI_variable(name, guid, origdata, len(origdata), attrs)
            if self.diff_var(self._uefi.get_EFI_variable(name, guid), origdata):
                nameguid = name + ' (' + guid + ')'
                self.logger.log_bad(f'RECOVERY FAILED. Variable {nameguid} remains corrupted. Original data value: {origdata}')
        return ret

    def check_vars(self, do_modify: bool, attribute: Optional[int] = None) -> int:
        res = ModuleResult.PASSED
        vars = self._uefi.list_EFI_variables()
        if vars is None:
            self.logger.log_warning('Could not enumerate UEFI Variables from runtime.')
            self.logger.log_important("Note that UEFI variables may still exist, OS just did not expose runtime UEFI Variable API to read them.\nYou can extract variables directly from ROM file via 'chipsec_util.py uefi nvram bios.bin' command and verify their attributes manually.")
            self.result.setStatusBit(self.result.status.VERIFY)
            return ModuleResult.WARNING

        uefispec_concern = []
        ro_concern = []
        rw_variables = []

        self.logger.log('[*] Testing UEFI variables ..')
        for name in vars.keys():
            if name is None:
                pass
            if vars[name] is None:
                pass

            if len(vars[name]) > 1:
                self.logger.log_important(f'Found two instances of the variable {name}.')
            for (off, buf, hdr, data, guid, attrs) in vars[name]:
                self.logger.log(f'[*] Variable {name} ({get_attr_string(attrs)}) Guid {guid} Size {hex(len(data))} ')
                perms = self.uefispec_vars.get(name)
                if perms is not None:
                    if perms != attrs:
                        attr_diffs = (perms ^ attrs)
                        extra_attr = attr_diffs & attrs
                        missing_attr = attr_diffs & ~extra_attr
                        uefispec_concern.append(name)
                        if extra_attr != 0:
                            self.logger.log_important('  Extra attributes:' + get_attr_string(extra_attr))
                            if (extra_attr & ~(EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS | EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS | EFI_VARIABLE_APPEND_WRITE) != 0):
                                res = ModuleResult.FAILED
                        if missing_attr != 0:
                            self.logger.log_important('  Missing attributes:' + get_attr_string(missing_attr))
                        if res != ModuleResult.FAILED:
                            res = ModuleResult.WARNING
                        self.result.setStatusBit(self.result.status.VERIFY)

                if do_modify:
                    if attribute != None:
                        attrs = attribute

                    self.logger.log(f"[*] Testing modification of {name} with attribute {str(attrs)}..")

                    if name in self.uefispec_ro_vars:
                        if self.can_modify(name, guid, data, attrs):
                            ro_concern.append(name)
                            self.logger.log_bad(f"Variable {name} should be read only.")
                            self.result.setStatusBit(self.result.status.POTENTIALLY_VULNERABLE)
                            res = ModuleResult.FAILED
                    else:
                        if self.can_modify(name, guid, data, attrs):
                            rw_variables.append(name)

        if uefispec_concern:
            self.logger.log('')
            self.logger.log_bad('Variables with attributes that differ from UEFI spec:')
            for name in uefispec_concern:
                self.logger.log(f'    {name}')

        if do_modify:
            if ro_concern:
                self.logger.log('')
                self.logger.log_bad('Variables that should have been read-only and were not:')
                for name in ro_concern:
                    self.logger.log(f'    {name}')

            if rw_variables:
                self.logger.log('')
                self.logger.log_unknown('Variables that are read-write (manual investigation is required):')
                for name in rw_variables:
                    self.logger.log(f'    {name}')

        self.logger.log('')

        if ModuleResult.PASSED == res:
            self.logger.log_passed('All checked EFI variables are protected according to spec.')
        elif ModuleResult.FAILED == res:
            self.logger.log_failed('Some EFI variables were not protected according to spec.')
        return res

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test("Access Control of EFI Variables")

        do_modify = (len(module_argv) > 0 and module_argv[0] == OPT_MODIFY)
        attribute = None

        if len(module_argv) > 1:
            attribute = int(module_argv[1], base=0)

        self.res = self.check_vars(do_modify, attribute)

        return self.result.getReturnCode(self.res)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2020, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Simple CPUID VMM emulation fuzzer

Usage:
    ``chipsec_main.py -i -m tools.vmm.cpuid_fuzz [-a random]``

    - ``random`` : Fuzz in random order (default is sequential)

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -i -m tools.vmm.cpuid_fuzz
    >>> chipsec_main.py -i -m tools.vmm.cpuid_fuzz -l log.txt
    >>> chipsec_main.py -i -m tools.vmm.cpuid_fuzz -a random

Additional options set within the module:
    - ``_NO_EAX_TO_FUZZ``        : No of EAX values to fuzz within each step
    - ``_EAX_FUZZ_STEP``         : Step to fuzz range of EAX values
    - ``_NO_ITERATIONS_TO_FUZZ`` : Number of iterations if `random` chosen
    - ``_FUZZ_ECX_RANDOM``       : Fuzz ECX with random values?
    - ``_MAX_ECX``               : Max ECX value
    - ``_EXCLUDE_CPUID``         : Exclude the following EAX values from fuzzing
    - ``_FLUSH_LOG_EACH_ITER``   : Flush log file after each iteration
    - ``_LOG_OUT_RESULTS``       : Log output results

.. note::
    - Returns a Warning by default
    - System may be in an unknown state, further evaluation may be needed

.. important::
    - This module is designed to run in a VM environment
    - Behavior on physical HW is undefined

"""

import random
from chipsec.module_common import BaseModule
from chipsec.library.returncode import ModuleResult

_MODULE_NAME = 'cpuid_fuzz'

#
# We will only be fuzzing _NO_EAX_TO_FUZZ range of EAX values each _EAX_FUZZ_STEP step
#
_NO_EAX_TO_FUZZ = 0x100
_EAX_FUZZ_STEP = 0x1000000

# Number of iterations if value is Randomly chosen
_NO_ITERATIONS_TO_FUZZ = 0x1000000

# Control values to be passed in ECX
_FUZZ_ECX_RANDOM = False
# Max value of ECX when fuzzed sequentially
_MAX_ECX = 0x100

# Exclude CPUID EAX which cause VM hang/crash
_EXCLUDE_CPUID = []

# Flush log file before each fuzz iteration
_FLUSH_LOG_EACH_ITER = False
# Log values of EAX, EBX, ECX, EDX which CPUID returns
_LOG_OUT_RESULTS = False


class cpuid_fuzz (BaseModule):
    def __init__(self):
        BaseModule.__init__(self)

    def fuzz_CPUID(self, eax_start, random_order=False):
        eax_range = _NO_EAX_TO_FUZZ
        eax_end = eax_start + eax_range
        self.logger.log(f'[*] Fuzzing CPUID with EAX in range 0x{eax_start:08X}:0x{eax_end:08X}..')
        it = 0
        if random_order:
            it_max = _NO_ITERATIONS_TO_FUZZ
        else:
            it_max = eax_range

        while it < it_max:
            if random_order:
                eax = random.randint(eax_start, eax_end)
            else:
                eax = eax_start + it
            if _FLUSH_LOG_EACH_ITER:
                self.logger.flush()
            if eax not in _EXCLUDE_CPUID:
                self.logger.log(f'[*] CPUID EAX: 0x{eax:08X}')
                if _FUZZ_ECX_RANDOM:
                    ecx = random.randint(0, 0xFFFFFFFF)
                    (r_eax, r_ebx, r_ecx, r_edx) = self.cs.cpu.cpuid(eax, ecx)
                else:
                    for ecx in range(_MAX_ECX):
                        self.logger.log(f'  > ECX: 0x{ecx:08X}')
                        if _FLUSH_LOG_EACH_ITER:
                            self.logger.flush()
                        (r_eax, r_ebx, r_ecx, r_edx) = self.cs.cpu.cpuid(eax, ecx)
                        if _LOG_OUT_RESULTS:
                            self.logger.log(f'    Out: EAX=0x{r_eax:08X}, EBX=0x{r_ebx:08X}, ECX=0x{r_ecx:08X}, EDX=0x{r_edx:08X}')
            it += 1
        return True

    def run(self, module_argv):
        self.logger.start_test('CPUID Fuzzer')

        _random_order = False
        if (len(module_argv) > 0) and ('random' == module_argv[0]):
            _random_order = True

        self.logger.log(f'[*] Configuration:')
        self.logger.log(f'    Mode: {"random" if _random_order else "sequential"}')
        self.logger.log(f'    Step to fuzz range of EAX values (_EAX_FUZZ_STEP): 0x{_EAX_FUZZ_STEP:X}')
        self.logger.log(f'    No of EAX values to fuzz within each step (_NO_EAX_TO_FUZZ): 0x{_NO_EAX_TO_FUZZ:X}')
        self.logger.log(f'    Fuzz ECX with random values? (_FUZZ_ECX_RANDOM): {_FUZZ_ECX_RANDOM:d}')
        self.logger.log(f'    Max ECX value (_MAX_ECX): 0x{_MAX_ECX:08X}')
        self.logger.log(f'    Exclude the following EAX values from fuzzing (_EXCLUDE_CPUID): {str(_EXCLUDE_CPUID)}')
        self.logger.log(f'    Flush log file after each iteration (_FLUSH_LOG_EACH_ITER): {_FLUSH_LOG_EACH_ITER:d}')
        self.logger.log(f'    Log output results (_LOG_OUT_RESULTS): {_LOG_OUT_RESULTS:d}')

        steps = 0x100000000 // _EAX_FUZZ_STEP
        for s in range(steps):
            self.fuzz_CPUID(s * _EAX_FUZZ_STEP, _random_order)

        self.logger.log_information('Module completed')
        self.logger.log_warning('System may be in an unknown state, further evaluation may be needed.')
        self.result.setStatusBit(self.result.status.VERIFY)
        self.res = self.result.getReturnCode(ModuleResult.WARNING)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


from unittest.mock import Mock

import chipsec.helper.replay.replayhelper as rph
from chipsec_main import ChipsecMain, parse_args
import chipsec.chipset as cs
import chipsec.library.logger


def run_chipsec_module(csm: ChipsecMain, module_replay_file: str) -> int:
    csm._cs.init(csm._platform, csm._pch, csm._helper, not csm._no_driver, csm._load_config, csm._ignore_platform)
    if module_replay_file:
        csm._helper.config_file = module_replay_file
        csm._helper._load()
    csm.load_module(csm._module, csm._module_argv)
    ret = csm.run_loaded_modules()
    return ret


def setup_run_destroy_module_with_mock_logger(init_replay_file: str, module_str: str, module_args: str = "", module_replay_file: str = "") -> int:
    chipsec.library.logger._logger.remove_chipsec_logger()
    chipsec.library.logger._logger = Mock()
    chipsec.library.logger._logger.VERBOSE = False
    chipsec.library.logger._logger.DEBUG = False
    chipsec.library.logger._logger.HAL = False
    retval = setup_run_destroy_module(init_replay_file, module_str, module_args, module_replay_file)
    chipsec.library.logger._logger = chipsec.library.logger.Logger()
    return retval


def setup_run_destroy_module(init_replay_file: str, module_str: str, module_args: str = "", module_replay_file: str = "") -> int:
    arg_str = f" {module_args}" if module_args else ""
    cli_cmds = f"-m {module_str}{arg_str}".strip().split(' ')
    cs._chipset = None
    par = parse_args(cli_cmds)
    csm = ChipsecMain(par, cli_cmds)
    replayHelper = rph.ReplayHelper(init_replay_file)
    csm._helper = replayHelper
    return run_chipsec_module(csm, module_replay_file)



#!/usr/bin/env python3
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Main application logic and automation functions
"""

# These are for debugging imports
import sys
import fnmatch
import argparse
import os
import re
import platform

import time
import traceback
from collections import OrderedDict
from typing import Dict, Any, Optional, Sequence

import chipsec.library.file
import chipsec.module
import chipsec.library.result_deltas
from chipsec.library import defines
from chipsec.library.returncode import ModuleResult, getModuleResultName
from chipsec import chipset
from chipsec.helper.oshelper import helper
from chipsec.library.logger import logger
from chipsec.library.banner import print_banner, print_banner_properties
from chipsec.testcase import ExitCode, TestCase, ReturnCodeResults, LegacyResults
from chipsec.library.exceptions import UnknownChipsetError, OsHelperError
from chipsec.library.options import Options
from chipsec.library.module_helper import enumerate_modules, print_modules

try:
    import importlib
except ImportError:
    pass


def parse_args(argv: Sequence[str]) -> Optional[Dict[str, Any]]:
    options = Options()

    default_helper = options.get_section_data('Main_Config', 'default_helper', None)
    default_exclude = options.get_list_data('Main_Config', 'exclude_modules', [])

    """Parse the arguments provided on the command line."""
    parser = argparse.ArgumentParser(usage='%(prog)s [options]', formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog=ExitCode.help_epilog, add_help=False)
    options = parser.add_argument_group('Options')
    options.add_argument('-h', '--help', help="Show this message and exit", action='store_true')
    options.add_argument('-m', '--module', dest='_module', help='Specify module to run (example: -m common.bios_wp)')
    options.add_argument('-mx', '--module_exclude', dest='_module_exclude', nargs='+', default=default_exclude, help='Specify module(s) to NOT run (example: -mx common.bios_wp common.cpu.cpu_info)')
    options.add_argument('-a', '--module_args', nargs='*', dest="_module_argv", help="Additional module arguments")
    options.add_argument('-v', '--verbose', help='Verbose logging', action='store_true')
    options.add_argument('--hal', help='HAL logging', action='store_true')
    options.add_argument('-d', '--debug', help='Debug logging', action='store_true')
    options.add_argument('-l', '--log', help='Output to log file')
    options.add_argument('-vv', '--vverbose', help='Very verbose logging (verbose + HAL + debug)', action='store_true')
    adv_options = parser.add_argument_group('Advanced Options')
    adv_options.add_argument('-p', '--platform', dest='_platform', help='Explicitly specify platform code',
                             choices=chipset.cs().Cfg.proc_codes, type=str.upper)
    adv_options.add_argument('--pch', dest='_pch', help='Explicitly specify PCH code', choices=chipset.cs().Cfg.pch_codes, type=str.upper)
    adv_options.add_argument('-n', '--no_driver', dest='_no_driver', action='store_true',
                             help="Chipsec won't need kernel mode functions so don't load chipsec driver")
    adv_options.add_argument('-i', '--ignore_platform', dest='_ignore_platform', action='store_true',
                             help='Run chipsec even if the platform is not recognized (Deprecated)')
    adv_options.add_argument('--csv', dest='_csv_out', help='Specify filename for CSV output')
    adv_options.add_argument('-j', '--json', dest='_json_out', help='Specify filename for JSON output')
    adv_options.add_argument('-x', '--xml', dest='_xml_out', help='Specify filename for xml output (JUnit style)')
    adv_options.add_argument('-k', '--markdown', dest='_markdown_out', help='Specify filename for markdown output')
    adv_options.add_argument('-t', '--moduletype', dest='USER_MODULE_TAGS', type=str.upper, default=[], help='Run tests of a specific type (tag)')
    adv_options.add_argument('--list_tags', dest='_list_tags', action='store_true', help='List all the available options for -t,--moduletype')
    adv_options.add_argument('-lm', '--list_modules', dest='_list_modules', action='store_true', help='List all the available options for -m,--module/-mx,--module_exclude')
    adv_options.add_argument('-I', '--include', dest='IMPORT_PATHS', default=[], help='Specify additional path to load modules from')
    adv_options.add_argument('--failfast', help="Fail on any exception and exit (don't mask exceptions)", action='store_true')
    adv_options.add_argument('--no_time', help="Don't log timestamps", action='store_true')
    adv_options.add_argument('--deltas', dest='_deltas_file', help='Specifies a JSON log file to compute result deltas from')
    adv_options.add_argument('--helper', dest='_helper', help='Specify OS Helper', choices=helper().get_available_helpers(), default=default_helper)
    adv_options.add_argument('-nb', '--no_banner', dest='_show_banner', action='store_false', help="Chipsec won't display banner information")
    adv_options.add_argument('--skip_config', dest='_load_config', action='store_false', help='Skip configuration and driver loading')
    adv_options.add_argument('-nl', dest='_autolog_disable', action='store_true', help="Chipsec won't save logs automatically")
    adv_options.add_argument('-rc', dest='_return_codes', help='Return codes mode', action='store_true')

    par = vars(parser.parse_args(argv))
    if par['help']:
        if par['_show_banner']:
            print_banner(argv, defines.get_version(), defines.get_message())
        parser.print_help()
        return None
    elif par['_list_modules']:
        print_modules(enumerate_modules())
        return None
    else:
        return par


class ChipsecMain:

    def __init__(self, switches, argv):
        self.logger = logger()
        self.CHIPSEC_FOLDER = chipsec.library.file.get_main_dir()
        self.PYTHON_64_BITS = True if (sys.maxsize > 2**32) else False
        self.Import_Path = "chipsec.modules."
        self.Modules_Path = chipsec.library.file.get_module_dir()
        self.Loaded_Modules = []
        self.AVAILABLE_TAGS = []
        self.MODPATH_RE = re.compile(r"^\w+(\.\w+)*$")
        self.version = defines.get_version()
        self.message = defines.get_message()
        self.__dict__.update(switches)
        self.argv = argv
        self._cs = chipset.cs()
        self.parse_switches()

    ##################################################################################
    # Module API
    ##################################################################################

    def import_module(self, module_path):
        module = None
        if not self.MODPATH_RE.match(module_path):
            self.logger.log_error(f'Invalid module path: {module_path}')
        else:
            try:
                module = importlib.import_module(module_path)
            except BaseException as msg:
                self.logger.log_error(f'Exception occurred during import of {module_path}: "{str(msg)}"')
                if self.logger.DEBUG:
                    self.logger.log_bad(traceback.format_exc())
                if self.failfast:
                    raise msg
        return module

    def verify_module_tags(self, module):
        run_it = True
        is_archived = False
        module_tags, metadata_tags = module.get_tags()
        if len(metadata_tags) > 0:
            self.logger.log(f'[*] Metadata tags: {metadata_tags}')
            if 'ARCHIVED' in metadata_tags:
                run_it = False
                is_archived = True
        if len(self.USER_MODULE_TAGS) > 0 or self._list_tags:
            run_it = False
            for mt in module_tags:
                if self._list_tags:
                    if mt not in self.AVAILABLE_TAGS:
                        self.AVAILABLE_TAGS.append(mt)
                elif mt in self.USER_MODULE_TAGS:
                    run_it = True
        return run_it, is_archived

    ##
    # full_path can be one of three things:
    # 1. the actual full path to the py or pyc file  i.e. c:\some_path\chipsec\modules\common\bios_wp.py
    # 2. a path to the pyc file inside a zip file    i.e. chipsec/modules/common/bios_wp.pyc
    # 3. the name of the module                      i.e. chipsec.modules.common.bios_wp
    def get_module_name(self, full_path):
        name = full_path.lower()
        # case #1, the full path: remove prefix
        if full_path.startswith(self.CHIPSEC_FOLDER + os.path.sep):
            name = full_path.replace(self.CHIPSEC_FOLDER + os.path.sep, '')
        else:
            for path in self.IMPORT_PATHS:
                if full_path.startswith(os.path.abspath(path) + os.path.sep):
                    name = full_path.replace(os.path.abspath(path) + os.path.sep, '')
        # case #1 and #2: remove the extension
        if name.lower().endswith('.py'):
            name = name[:-3]
        if name.lower().endswith('.pyc'):
            name = name[:-4]
        # case #1: replace slashes with dots
        name = name.replace(os.path.sep, '.')
        # case #2: when in a zip it is always forward slash
        name = name.replace('/', '.')

        # Add 'chipsec.modules.' if shor module name was provided and alternative import paths were not specified
        if [] == self.IMPORT_PATHS and not name.startswith(self.Import_Path):
            name = self.Import_Path + name

        return name

    #
    # module_path is a file path relative to chipsec
    # E.g. chipsec/modules/common/module.py
    #
    def load_module(self, module_path, module_argv):
        module_name = self.get_module_name(module_path)
        module = chipsec.module.Module(module_name)

        if module not in self.Loaded_Modules:
            if self._module_exclude:
                if not [i for i in self._module_exclude if i in module.name]:
                    self.Loaded_Modules.append((module, module_argv))
            else:
                self.Loaded_Modules.append((module, module_argv))
        return True

    def load_modules_from_path(self, from_path, recursive=True):
        if self.logger.DEBUG:
            self.logger.log(f'[*] Path: {os.path.abspath(from_path)}')
        for dirname, subdirs, mod_fnames in os.walk(os.path.abspath(from_path)):
            if not recursive:
                while len(subdirs) > 0:
                    subdirs.pop()
            for modx in mod_fnames:
                if fnmatch.fnmatch(modx, '*.py') and not fnmatch.fnmatch(modx, '__init__.py') and not fnmatch.fnmatch(modx, '*sidekick.py'):
                    self.load_module(os.path.join(dirname, modx), self._module_argv)
        self.Loaded_Modules.sort()

    def load_my_modules(self):
        #
        # Step 1.
        # Load modules common to all supported platforms
        #
        common_path = os.path.join(self.Modules_Path, 'common')
        self.logger.log(f'[*] loading common modules from "{common_path.replace(os.getcwd(), ".")}" ..')
        self.load_modules_from_path(common_path)
        #
        # Step 2.
        # Load platform-specific modules from the corresponding platform module directory
        #
        chipset_path = os.path.join(self.Modules_Path, self._cs.Cfg.code.lower())
        if (chipset.CHIPSET_CODE_UNKNOWN != self._cs.Cfg.code) and os.path.exists(chipset_path):
            self.logger.log(f'[*] loading platform specific modules from \"{chipset_path.replace(os.getcwd(), ".")}\" ..')
            self.load_modules_from_path(chipset_path)
        else:
            self.logger.log("[*] No platform specific modules to load")
        #
        # Step 3.
        # Enumerate all modules from the root module directory
        #
        self.logger.log(f'[*] loading modules from \"{self.Modules_Path.replace(os.getcwd(), ".")}\" ..')
        self.load_modules_from_path(self.Modules_Path, False)

    def load_user_modules(self):
        for import_path in self.IMPORT_PATHS:
            self.logger.log(f'[*] loading modules from \"{import_path}\" ..')
            self.load_modules_from_path(import_path)

    def clear_loaded_modules(self):
        del self.Loaded_Modules[:]

    def print_loaded_modules(self):
        if self.Loaded_Modules == []:
            self.logger.log("No modules have been loaded")
        for (modx, _) in self.Loaded_Modules:
            self.logger.log(f'[+] loaded {modx}')

    def run_module(self, modx, module_argv):
        result = None
        try:
            if not modx.do_import():
                return ModuleResult.ERROR
            if self.logger.DEBUG and not self._list_tags:
                self.logger.log(f'[*] Module path: {modx.get_location()}')

            run_it, is_archived = self.verify_module_tags(modx)
            if run_it:
                result = modx.run(module_argv)
            else:
                modx.get_module_object()
                if is_archived:
                    modx.mod_obj.result.setStatusBit(modx.mod_obj.result.status.ARCHIVED)
                    res = ModuleResult.ARCHIVED
                else:
                    modx.mod_obj.result.setStatusBit(modx.mod_obj.result.status.NOT_APPLICABLE)
                    res = ModuleResult.NOTAPPLICABLE
                return modx.mod_obj.result.getReturnCode(res, False)
        except BaseException as msg:
            if self.logger.DEBUG:
                self.logger.log_bad(traceback.format_exc())
            self.logger.log_error(f'Exception occurred during {modx.get_name()}.run(): \'{str(msg)}\'')
            raise msg
        return result

    def run_loaded_modules(self):
        results = ReturnCodeResults() if self._return_codes else LegacyResults()
        results.add_properties(self.properties())

        # Print a list of all loaded modules
        self.print_loaded_modules()
        if not self._list_tags:
            self.logger.log("[*] running loaded modules ..")

        t = time.time()
        for (modx, modx_argv) in self.Loaded_Modules:
            test_result = TestCase(modx.get_name())
            test_result.start_module()

            # Run the module
            try:
                result = self.run_module(modx, modx_argv)
            except BaseException:
                results.add_exception(modx)
                result = ModuleResult.ERROR
                if self.logger.DEBUG:
                    self.logger.log_bad(traceback.format_exc())
                if self.failfast:
                    raise

            # Populate results
            test_result.end_module(getModuleResultName(result, self._return_codes), result, modx_argv if modx_argv else None)
            results.add_testcase(test_result)

        runtime = time.time() - t if not self.no_time else None

        if self._json_out:
            chipsec.library.file.write_file(self._json_out, results.json_full())

        if self._xml_out:
            chipsec.library.file.write_file(self._xml_out, results.xml_full(self._xml_out, runtime))

        if self._markdown_out:
            chipsec.library.file.write_file(self._markdown_out, results.markdown_full(self._markdown_out))

        if self._csv_out:
            self.logger.log_csv(self._csv_out, results.test_cases)

        test_deltas = None
        if self._deltas_file is not None:
            prev_results = chipsec.library.result_deltas.get_json_results(self._deltas_file)
            if prev_results is None:
                self.logger.log_error("Delta processing disabled.  Displaying results summary.")
            else:
                test_deltas = chipsec.library.result_deltas.compute_result_deltas(prev_results, results.get_results())
                chipsec.library.result_deltas.display_deltas(test_deltas, self.no_time, t)
        elif not self._list_tags and results.get_current_testcase is not None:
            results.print_summary(runtime)
        else:
            self.logger.log("[*] Available tags are:")
            for at in self.AVAILABLE_TAGS:
                self.logger.log(f'    {at}')

        return results.get_return_code()

    ##################################################################################
    # Running all relevant modules
    ##################################################################################

    def run_all_modules(self):
        self.load_my_modules()
        self.load_user_modules()

        return self.run_loaded_modules()

    def parse_switches(self) -> None:
        """Configure logging parameters based on arguments"""
        self.logger.set_log_level(self.verbose, self.hal, self.debug, self.vverbose)
        if self.log:
            self.logger.set_log_file(self.log)
            self._autolog_disable = True
        if self._autolog_disable is False:
            self.logger.set_autolog_file("main")
        if self._return_codes:
            self.logger.log_warning("Return codes feature is currently Work in Progress!!!")
            self._cs.using_return_codes = True
        if self._module_argv and len(self._module_argv) == 1 and self._module_argv[0].count(','):
            self.logger.log("[*] Use of the -a command no longer needs to have arguments concatenated with ','")
            self._module_argv = self._module_argv[0].split(',')
        if self._ignore_platform:
            self.logger.log_warning("Ignoring unsupported platform warning and continue execution.")
            self.logger.log_warning("Most results cannot be trusted.")
            self.logger.log_warning("Unless a platform independent module is being run, do not file issues against this run.")

    def properties(self):
        ret = OrderedDict()
        ret["OS"] = f'{self._cs.helper.os_system} {self._cs.helper.os_release} {self._cs.helper.os_version} {self._cs.helper.os_machine}'
        ret["Python"] = f'Python {platform.python_version()}'
        ret["Platform"] = f'{self._cs.Cfg.longname}, CPUID: {self._cs.Cfg.cpuid}, VID: {self._cs.Cfg.vid:04X}, DID: {self._cs.Cfg.did:04X}, RID: {self._cs.Cfg.rid:02X}'
        if not self._cs.is_atom():
            ret["PCH"] = f'{self._cs.Cfg.pch_longname}, VID: {self._cs.Cfg.pch_vid:04X}, DID: {self._cs.Cfg.pch_did:04X} RID: {self._cs.Cfg.pch_rid:02X}'
        ret["Version"] = f'{self.version}'
        ret["Message"] = f'{self.message}'
        return ret

    ##################################################################################
    # Entry point for command-line execution
    ##################################################################################

    def main(self) -> int:

        if self._show_banner:
            print_banner(self.argv, defines.get_version(), defines.get_message())

        for import_path in self.IMPORT_PATHS:
            sys.path.append(os.path.abspath(import_path))

        try:
            self._cs.init(self._platform, self._pch, self._helper, not self._no_driver, self._load_config, self._ignore_platform)
        except UnknownChipsetError as msg:
            self.logger.log_error(f'Platform is not supported ({str(msg)}).')
            if self._ignore_platform:
                self.logger.log_error('To specify a cpu please use -p command-line option')
                self.logger.log_error('To specify a pch please use --pch command-line option\n')
                self.logger.log_error('If the correct configuration is not loaded, results should not be trusted.')
                if self.logger.DEBUG:
                    self.logger.log_bad(traceback.format_exc())
                if self.failfast:
                    raise msg
                return ExitCode.EXCEPTION
            self.logger.log_warning("Platform dependent functionality is likely to be incorrect")
        except OsHelperError as os_helper_error:
            self.logger.log_error(str(os_helper_error))
            if self.logger.DEBUG:
                self.logger.log_bad(traceback.format_exc())
            if self.failfast:
                raise os_helper_error
            return ExitCode.EXCEPTION
        except BaseException as be:
            self.logger.log_bad(traceback.format_exc())
            if self.failfast:
                raise be
            return ExitCode.EXCEPTION

        if self._show_banner:
            print_banner_properties(self._cs, defines.os_version())

        self.logger.log(" ")

        if self.logger.DEBUG:
            self.logger.log(f'[*] Running from {os.getcwd()}')

        self.main_return = 0
        if self._module:
            self.load_module(self._module, self._module_argv)
            self.main_return = self.run_loaded_modules()
        else:
            self.main_return = self.run_all_modules()
        if not self._no_driver:
            self._cs.destroy_helper()
        del self._cs
        self.logger.disable()
        return self.main_return


def run(cli_cmd: str = '') -> int:
    cli_cmds = []
    if cli_cmd:
        cli_cmds = cli_cmd.strip().split(' ')
    return main(cli_cmds)


def main(argv: Sequence[str] = sys.argv[1:]) -> int:
    par = parse_args(argv)
    if par is not None:
        chipsecMain = ChipsecMain(par, argv)
        return chipsecMain.main()
    return ExitCode.OK


if __name__ == "__main__":
    sys.exit(main())



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

import unittest

from chipsec.hal.acpi_tables import APIC, BGRT, DMAR, XSDT, UEFI_TABLE, WSMT


class TestACPITables(unittest.TestCase):
    """Test the ACPI Table Structures and Parsing."""

    def test_verify_dmar_table_format(self):
        test_dmar = DMAR()
        value_re = r'^=?[BHQsI\d]*'
        key_re = r'^\w+_FORMAT'
        self.assertEqual(type(test_dmar.DMAR_TABLE_FORMAT), dict)
        for key, value in test_dmar.DMAR_TABLE_FORMAT.items():
            self.assertRegex(key, key_re)
            self.assertRegex(value, value_re)

    def test_dmar_parse(self):
        test_table_content = b'&\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\xd9\xfe\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00\x00\x00 \x00\x01\x00\x00\x00\x00\x10\xd9\xfe\x00\x00\x00\x00\x03\x08\x00\x00\x02\xf0\x1f\x00\x04\x08\x00\x00\x00\x00\x1f\x00\x01\x00 \x00\x00\x00\x00\x00\x00\xa0\xb9z\x00\x00\x00\x00\xff?\xdez\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x14\x00\x01\x00 \x00\x00\x00\x00\x00\x00\x00\x80{\x00\x00\x00\x00\xff\xff\xff\x7f\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00'
        test_dmar = DMAR()
        test_dmar.parse(test_table_content)
        self.assertEqual(len(test_dmar.dmar_structures), 4)

    def test__str__(self):
        test_table_content = b'&\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\xd9\xfe\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00\x00\x00 \x00\x01\x00\x00\x00\x00\x10\xd9\xfe\x00\x00\x00\x00\x03\x08\x00\x00\x02\xf0\x1f\x00\x04\x08\x00\x00\x00\x00\x1f\x00\x01\x00 \x00\x00\x00\x00\x00\x00\xa0\xb9z\x00\x00\x00\x00\xff?\xdez\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x14\x00\x01\x00 \x00\x00\x00\x00\x00\x00\x00\x80{\x00\x00\x00\x00\xff\xff\xff\x7f\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00'
        # test_table_content = b'&\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\xd9\xfe\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00`\xd8\xfe\x00\x00\x00\x00\x02\x08\x00\x00\x00\x00\x07\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00p\xd8\xfe\x00\x00\x00\x00\x02\x08\x00\x00\x00\x00\x07\x03\x00\x00 \x00\x01\x00\x00\x00\x00\x10\xd9\xfe\x00\x00\x00\x00\x03\x08\x00\x00\x02\x00\x1e\x07\x04\x08\x00\x00\x00\x00\x1e\x06\x01\x00 \x00\x00\x00\x00\x00\x00\x00\x00/\x00\x00\x00\x00\xff\xff\x7fO\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00'
        test_dmar = DMAR()
        test_dmar.parse(test_table_content)
        str_header = 'DMAR Table Contents'
        str_contents = test_dmar.__str__()
        self.assertIn(str_header, str_contents)
        self.assertTrue(1500 <= len(str_contents) <= 2000)

    def test_get_structure_dmar_type_7(self):
        test_dmar = DMAR()
        type = 0x07
        structure_DMAR_ret = test_dmar._get_structure_DMAR(type, b'\x00\x00\x00\x00')
        self.assertIn(f"\n  Unknown DMAR structure 0x{type:02X}\n", structure_DMAR_ret)

    def test_verify_apic_table_format(self):
        test_apic = APIC()
        value_re = r'^[<=]?[BHQsI\d]*'
        key_re = r'^\w+'
        self.assertEqual(type(test_apic.APIC_TABLE_FORMAT), dict)
        for key, value in test_apic.APIC_TABLE_FORMAT.items():
            self.assertRegex(key, key_re)
            self.assertRegex(value, value_re)

    def test_apic_parse(self):
        test_table_content = b'\x00\x00\xe0\xfe\x01\x00\x00\x00\x00\x08\x01\x00\x01\x00\x00\x00\x04\x06\x01\x05\x00\x01\x00\x08\x02\x02\x01\x00\x00\x00\x04\x06\x02\x05\x00\x01\x00\x08\x03\x01\x01\x00\x00\x00\x04\x06\x03\x05\x00\x01\x00\x08\x04\x03\x01\x00\x00\x00\x04\x06\x04\x05\x00\x01\x01\x0c\x02\x00\x00\x00\xc0\xfe\x00\x00\x00\x00\x02\n\x00\x00\x02\x00\x00\x00\x00\x00\x02\n\x00\t\t\x00\x00\x00\r\x00'
        test_apic = APIC()
        test_apic.parse(test_table_content)
        self.assertEqual(len(test_apic.apic_structs), 11)

    def test_xsdt_parse(self):
        test_table_content = b'h\x9daz\x00\x00\x00\x00\x80\x9eaz\x00\x00\x00\x00\x08\x9faz\x00\x00\x00\x00P\x9faz\x00\x00\x00\x00\xf0\x9faz\x00\x00\x00\x000\xa0az\x00\x00\x00\x00\x90\xa3az\x00\x00\x00\x00\xf8\xd4az\x00\x00\x00\x000\xd5az\x00\x00\x00\x00\x18\xdfaz\x00\x00\x00\x00`\xdfaz\x00\x00\x00\x00\xc0\xe9az\x00\x00\x00\x00p\x01bz\x00\x00\x00\x00\x08\x02bz\x00\x00\x00\x00P\x03bz\x00\x00\x00\x00\xf0\x05bz\x00\x00\x00\x00\xf85bz\x00\x00\x00\x00\xd86bz\x00\x00\x00\x00\xc89bz\x00\x00\x00\x00\x00:bz\x00\x00\x00\x00X:bz\x00\x00\x00\x00@Mbz\x00\x00\x00\x00\xe8Mbz\x00\x00\x00\x00\x18Nbz\x00\x00\x00\x00PNbz\x00\x00\x00\x00\x88Nbz\x00\x00\x00\x00\xe0Nbz\x00\x00\x00\x00'
        test_xsdt = XSDT()
        test_xsdt.parse(test_table_content)
        self.assertEqual(len(test_xsdt.Entries), 27)

    def test_bgrt_parse(self):
        test_table_content = b"\x01\x00\x01\x00\x18P\x9fv\x00\x00\x00\x00\xf3\x02\x00\x00'\x01\x00\x00"
        test_bgrt = BGRT()
        test_bgrt.parse(test_table_content)
        self.assertEqual(test_bgrt.Version, 1)
        self.assertEqual(test_bgrt.Status, 1)
        self.assertEqual(test_bgrt.ImageOffsetY, 295)

    def test_uefi_parse(self):
        test_table_content = b'\xe2\xd8\x8e\xc6\xc6\x9d\xbdL\x9d\x94\xdbe\xac\xc5\xc328\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\xb0\x9fz\x00\x00\x00\x00'
        test_uefi = UEFI_TABLE()
        test_uefi.parse(test_table_content)
        self.assertEqual(test_uefi.smi, 1)
        self.assertEqual(test_uefi.buf_addr, 0x7a9fb00000000000)

    def test_wsmt_parse(self):
        test_table_content = b'\x00\x00\x00\x00'
        test_wsmt = WSMT()
        test_wsmt.parse(test_table_content)
        self.assertEqual(test_wsmt.fixed_comm_buffers, 0)
        self.assertEqual(test_wsmt.comm_buffer_nested_ptr_protection, 0)
        self.assertEqual(test_wsmt.system_resource_protection, 0)
    # def test_get_structure_dmar(self):
    #     test_table_content = b'&\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\xd9\xfe\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00`\xd8\xfe\x00\x00\x00\x00\x02\x08\x00\x00\x00\x00\x07\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00p\xd8\xfe\x00\x00\x00\x00\x02\x08\x00\x00\x00\x00\x07\x03\x00\x00 \x00\x01\x00\x00\x00\x00\x10\xd9\xfe\x00\x00\x00\x00\x03\x08\x00\x00\x02\x00\x1e\x07\x04\x08\x00\x00\x00\x00\x1e\x06\x01\x00 \x00\x00\x00\x00\x00\x00\x00\x00/\x00\x00\x00\x00\xff\xff\x7fO\x00\x00\x00\x00\x01\x08\x00\x00\x00\x00\x02\x00'
    #     test_dmar = DMAR()
    #     i = 1
    #     for key, value in test_dmar.DMAR_TABLE_FORMAT.items():
    #         size = struct.calcsize(value)
    #         test_dmar._get_structure_DMAR(i, test_table_content[])
    #         i += 1
    #     print(test_dmar._get_DMAR_structure_SATC(b'\x00'))


if __name__ == '__main__':
    unittest.main()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#


# To execute: python -m unittest tests.helpers.test_recordhelper
from os import remove
import unittest
from os.path import join, isfile
import chipsec.helper.replay.replayhelper as rph
import chipsec.helper.record.recordhelper as rch


class RecordHelperTest(unittest.TestCase):

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        if isfile(self.temp_file):
            remove(self.temp_file)

    def test_record(self):
        self.original_file = join("tests", "helpers", "kblreplaytest.json")
        self.temp_file = join("tests", "helpers", "kblrecordtest_temp.json")
        if isfile(self.temp_file):
            remove(self.temp_file)
        self.replayhelper = rph.ReplayHelper(self.original_file)
        self.recordhelper = rch.RecordHelper(self.temp_file)
        self.recordhelper.switch_subhelper(self.replayhelper)
        self.assertTrue(self.recordhelper.create())
        self.assertTrue(self.recordhelper.start())

        pci_read_value = self.recordhelper.read_pci_reg(0, 0, 0, 0, 0x1)
        self.assertEqual(pci_read_value, 0x86)

        pci_read_value = self.recordhelper.read_pci_reg(0, 0, 0, 0, 0x2)
        self.assertEqual(pci_read_value, 0x8086)

        pci_read_value = self.recordhelper.read_pci_reg(0, 0, 0, 0, 0x4)
        self.assertEqual(pci_read_value, 0x59048086)

        pci_read_value = self.recordhelper.read_pci_reg(0, 0x1f, 0, 0, 0x1)
        self.assertEqual(pci_read_value, 0x86)

        pci_read_value = self.recordhelper.read_pci_reg(0, 0x1f, 0, 0, 0x2)
        self.assertEqual(pci_read_value, 0x8086)

        pci_read_value = self.recordhelper.read_pci_reg(0, 0x1f, 0, 0, 0x4)
        self.assertEqual(pci_read_value, 0x9D4E8086)

        thread_count = self.recordhelper.get_threads_count()
        self.assertEqual(thread_count, 4)

        mem_value = self.recordhelper.read_phys_mem(0x5000, 0x2)
        self.assertEqual(mem_value, b'\xd1\x99')

        mem_value = self.recordhelper.read_phys_mem(0x5001, 0x2)
        self.assertEqual(mem_value, b'\x99\xaa')

        mem_value = self.recordhelper.read_phys_mem(0x5010, 0x2)
        self.assertEqual(mem_value, b'\xf3\x99')

        cpuid_value = self.recordhelper.cpuid(1, 0)
        self.assertEqual(cpuid_value, (526057, 51382272, 2147154879, 3219913727))

        cpuid_value = self.recordhelper.cpuid(2, 0)
        self.assertEqual(cpuid_value, (1979933441, 15775231, 0, 12779520))

        self.recordhelper.write_pci_reg(0, 0x3, 0, 0x2D, 0x33, 0x1)
        pci_read_value = self.recordhelper.read_pci_reg(0, 0x3, 0, 0x2d, 0x1)
        self.assertEqual(pci_read_value, 0xFF)

        self.recordhelper.write_pci_reg(0, 0x0, 0, 0x0, 0x44, 0x1)
        pci_read_value = self.recordhelper.read_pci_reg(0, 0x0, 0, 0x0, 0x2)
        self.assertEqual(pci_read_value, 32902)

        mmio_read_value = self.recordhelper.read_mmio_reg(0xfed0, 0x3)
        self.assertEqual(mmio_read_value, 0xababab)

        self.recordhelper.write_mmio_reg(0x123, 0x1, 0x22)
        mmio_read_value = self.recordhelper.read_mmio_reg(0x123, 0x1)
        self.assertEqual(mmio_read_value, 0x22)

        self.assertTrue(self.recordhelper.stop())
        self.assertTrue(self.recordhelper.delete())
        self.assertTrue(self._compare_replay_and_record())

    def _compare_replay_and_record(self):
        with open(self.original_file, 'r') as of:
            original_lines = of.readlines()
        with open(self.temp_file, 'r') as tf:
            temp_lines = tf.readlines()
        if len(original_lines) != len(temp_lines):
            return False
        for i in range(len(original_lines)):
            if original_lines[i] != temp_lines[i]:
                print(f"{self.original_file} line {i+1}:{original_lines[i]}")
                print(f"{self.temp_file} line {i+1}:{temp_lines[i]}")
                return False
        return True




# !/usr/bin/python
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util ec dump    [<size>]
>>> chipsec_util ec command <command>
>>> chipsec_util ec read    <offset> [<size>]
>>> chipsec_util ec write   <offset> <byte_val>
>>> chipsec_util ec index   [<offset>]

Examples:

>>> chipsec_util ec dump
>>> chipsec_util ec command 0x001
>>> chipsec_util ec read    0x2F
>>> chipsec_util ec write   0x2F 0x00
>>> chipsec_util ec index
"""

from argparse import ArgumentParser

from chipsec.command import BaseCommand, toLoad

from chipsec.library.logger import print_buffer_bytes
from chipsec.hal.ec import EC


# Embedded Controller
class ECCommand(BaseCommand):

    def requirements(self) -> toLoad:
        if hasattr(self, 'func'):
            return toLoad.Driver
        return toLoad.Nil

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=__doc__)

        parser_offset = ArgumentParser(add_help=False)
        parser_offset.add_argument('offset', type=lambda x: int(x, 0), nargs='?', default=0, help="offset")

        parser_dmpsz = ArgumentParser(add_help=False)
        parser_dmpsz.add_argument("size", type=lambda sz: int(sz, 0), nargs='?', default=0x160, help="size")

        parser_rdsz = ArgumentParser(add_help=False)
        parser_rdsz.add_argument("size", type=lambda sz: int(sz, 0), nargs='?', default=None, help="size")

        subparsers = parser.add_subparsers()

        parser_command = subparsers.add_parser('command')
        parser_command.add_argument("cmd", type=lambda sz: int(sz, 0), help="EC command to issue")
        parser_command.set_defaults(func=self.command)

        parser_dump = subparsers.add_parser('dump', parents=[parser_dmpsz])
        parser_dump.set_defaults(func=self.dump)

        parser_read = subparsers.add_parser('read', parents=[parser_offset, parser_rdsz])
        parser_read.set_defaults(func=self.read)

        parser_write = subparsers.add_parser('write', parents=[parser_offset])
        parser_write.add_argument("wval", type=lambda sz: int(sz, 0), help="byte value to write into EC memory")
        parser_write.set_defaults(func=self.write)

        parser_index = subparsers.add_parser('index', parents=[parser_offset])
        parser_index.set_defaults(func=self.index)

        parser.parse_args(self.argv, namespace=self)

    def set_up(self) -> None:
        self._ec = EC(self.cs)

    def dump(self) -> None:
        self.logger.log("[CHIPSEC] EC dump")

        buf = self._ec.read_range(0, self.size)
        print_buffer_bytes(buf)

    def command(self) -> None:
        self.logger.log(f'[CHIPSEC] Sending EC command 0x{self.cmd:X}')

        self._ec.write_command(self.cmd)

    def read(self) -> None:
        if self.size:
            buf = self._ec.read_range(self.offset, self.size)
            self.logger.log(f'[CHIPSEC] EC memory read: offset 0x{self.offset:X} size 0x{self.size:X}')
            print_buffer_bytes(buf)
        else:
            val = self._ec.read_memory(
                self.offset) if self.offset < 0x100 else self._ec.read_memory_extended(self.offset)
            self.logger.log(f'[CHIPSEC] EC memory read: offset 0x{self.offset:X} = 0x{val:X}')

    def write(self) -> None:
        self.logger.log(f'[CHIPSEC] EC memory write: offset 0x{self.offset:X} = 0x{self.wval:X}')

        if self.offset < 0x100:
            self._ec.write_memory(self.offset, self.wval)
        else:
            self._ec.write_memory_extended(self.offset, self.wval)

    def index(self) -> None:

        if self.offset:
            val = self._ec.read_idx(self.offset)
            self.logger.log(f'[CHIPSEC] EC index I/O: reading memory offset 0x{self.offset:X}: 0x{val:X}')
        else:
            self.logger.log("[CHIPSEC] EC index I/O: dumping memory...")
            mem = [self._ec.read_idx(off) for off in range(0x10000)]
            print_buffer_bytes(mem)


commands = {'ec': ECCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Check Memory Remapping Configuration

Reference:
    - `Preventing & Detecting Xen Hypervisor Subversions <http://www.invisiblethingslab.com/resources/bh08/part2-full.pdf>`_ by Joanna Rutkowska & Rafal Wojtczuk

Usage:
  ``chipsec_main -m common.remap``

Example:
    >>> chipsec_main.py -m common.remap

Registers used:
    - PCI0.0.0_REMAPBASE
    - PCI0.0.0_REMAPLIMIT
    - PCI0.0.0_REMAPBASEMC1
    - PCI0.0.0_REMAPLIMITMC1
    - PCI0.0.0_TOUUD
    - PCI0.0.0_TOLUD
    - PCI0.0.0_TSEGMB

.. note::
    - This module will only run on Core platforms.

"""

from chipsec.module_common import BaseModule, HWCONFIG, SMM
from chipsec.library.returncode import ModuleResult
from chipsec.library.defines import BIT32, ALIGNED_1MB
from chipsec.library.exceptions import RegisterNotFoundError
from typing import Tuple

_MODULE_NAME = 'remap'

TAGS = [SMM, HWCONFIG]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'REMAP']

_REMAP_ADDR_MASK = 0x7FFFF00000
_TOLUD_MASK = 0xFFFFF000


class remap(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)

    def is_supported(self) -> bool:
        if self.cs.is_core():
            rbase_exist = self.cs.register.is_defined('PCI0.0.0_REMAPBASE')
            rlimit_exist = self.cs.register.is_defined('PCI0.0.0_REMAPLIMIT')
            touud_exist = self.cs.register.is_defined('PCI0.0.0_TOUUD')
            tolud_exist = self.cs.register.is_defined('PCI0.0.0_TOLUD')
            tseg_exist = self.cs.register.is_defined('PCI0.0.0_TSEGMB')
            if rbase_exist and rlimit_exist and touud_exist and tolud_exist and tseg_exist:
                return True
            self.logger.log_important('Required register definitions not defined for platform.  Skipping module.')
        else:
            self.logger.log_important('Not a Core (client) platform.  Skipping module.')

        return False

    def is_ibecc_enabled(self) -> bool:
        if self.cs.register.is_defined('IBECC_ACTIVATE'):
            edsr = self.cs.register.read_field('IBECC_ACTIVATE', 'IBECC_EN')
            if edsr == 1:
                return True
            else:
                self.logger.log_verbose('IBECC is not enabled!')
        else:
            self.logger.log_verbose('IBECC is not defined!')
        return False

    def get_remap_registers(self, base_reg: str, limit_reg: str) -> Tuple[int, int]:
        if self.cs.register.is_defined(base_reg) and self.cs.register.is_defined(limit_reg):
            base = self.cs.register.read(base_reg)
            limit = self.cs.register.read(limit_reg)
            return base, limit
        raise RegisterNotFoundError('Required register definitions not defined for platform.')

    def check_remap_config(self) -> int:
        is_warning = False
        remapMC0Found = True
        remapMC1Found = True

        try:
            remapbase, remaplimit = self.get_remap_registers('PCI0.0.0_REMAPBASE', 'PCI0.0.0_REMAPLIMIT')
        except RegisterNotFoundError:
            remapbase, remaplimit = 0, 0
            remapMC0Found = False
        try:
            remapbasemc1, remaplimitmc1 = self.get_remap_registers('PCI0.0.0_REMAPBASEMC1', 'PCI0.0.0_REMAPLIMITMC1')
        except RegisterNotFoundError:
            remapMC1Found = False

        touud = self.cs.register.read('PCI0.0.0_TOUUD')
        tolud = self.cs.register.read('PCI0.0.0_TOLUD')
        tsegmb = self.cs.register.read('PCI0.0.0_TSEGMB')
        self.logger.log('[*] Registers:')
        self.logger.log(f'[*]   TOUUD     : 0x{touud:016X}')
        if remapMC0Found:
            self.logger.log(f'[*]   REMAPLIMIT: 0x{remaplimit:016X}')
            self.logger.log(f'[*]   REMAPBASE : 0x{remapbase:016X}')
        if remapMC1Found:
            self.logger.log(f'[*]   REMAPLIMITMC1: 0x{remaplimitmc1:016X}')
            self.logger.log(f'[*]   REMAPBASEMC1: 0x{remapbasemc1:016X}')
        self.logger.log(f'[*]   TOLUD     : 0x{tolud:08X}')
        self.logger.log(f'[*]   TSEGMB    : 0x{tsegmb:08X}')
        self.logger.log('')

        if remapMC0Found and self.cs.register.is_all_ffs('PCI0.0.0_REMAPBASE', remapbase) and self.cs.register.is_all_ffs('PCI0.0.0_REMAPLIMIT', remaplimit) and remapMC1Found:
            remapbase_register = 'PCI0.0.0_REMAPBASEMC1'
            remaplimit_register = 'PCI0.0.0_REMAPLIMITMC1'
            remapbase, remaplimit = remapbasemc1, remaplimitmc1
        else:
            remapbase_register = 'PCI0.0.0_REMAPBASE'
            remaplimit_register = 'PCI0.0.0_REMAPLIMIT'

        ia_untrusted = 0
        remapbase_lock = 0
        remaplimit_lock = 0
        if self.cs.register.has_field('MSR_BIOS_DONE', 'IA_UNTRUSTED'):
            ia_untrusted = self.cs.register.read_field('MSR_BIOS_DONE', 'IA_UNTRUSTED')
        if self.cs.register.has_field(remapbase_register, 'LOCK'):
            remapbase_lock = remapbase & 0x1
        if self.cs.register.has_field(remaplimit_register, 'LOCK'):
            remaplimit_lock = remaplimit & 0x1
        touud_lock = touud & 0x1
        tolud_lock = tolud & 0x1
        remapbase &= _REMAP_ADDR_MASK
        remaplimit &= _REMAP_ADDR_MASK
        touud &= _REMAP_ADDR_MASK
        tolud &= _TOLUD_MASK
        tsegmb &= _TOLUD_MASK
        self.logger.log('[*] Memory Map:')
        self.logger.log(f'[*]   Top Of Upper Memory: 0x{touud:016X}')
        self.logger.log(f'[*]   Remap Limit Address: 0x{(remaplimit | 0xFFFFF):016X}')
        self.logger.log(f'[*]   Remap Base Address : 0x{remapbase:016X}')
        self.logger.log(f'[*]   4GB                : 0x{BIT32:016X}')
        self.logger.log(f'[*]   Top Of Low Memory  : 0x{tolud:016X}')
        self.logger.log(f'[*]   TSEG (SMRAM) Base  : 0x{tsegmb:016X}')
        self.logger.log('')

        if self.cs.register.is_all_ffs(remapbase_register, remapbase) or self.cs.register.is_all_ffs(remaplimit_register, remaplimit):
            res = ModuleResult.FAILED
            self.result.setStatusBit(self.result.status.CONFIGURATION)
            self.logger.log('[!]   Memory Remap status is Unknown')
            return self.result.getReturnCode(res)

        remap_ok = True

        self.logger.log('[*] Checking memory remap configuration..')

        if remapbase == remaplimit:
            self.logger.log('[!]   Memory Remap status is Unknown')
            is_warning = True
        elif remapbase > remaplimit:
            self.logger.log('[*]   Memory Remap is disabled')
        else:
            self.logger.log('[*]   Memory Remap is enabled')
            remaplimit_addr = (remaplimit | 0xFFFFF)
            if self.is_ibecc_enabled():
                ok = (remaplimit_addr > touud) and (remapbase < touud)
            else:
                ok = ((remaplimit_addr + 1) == touud)
            remap_ok = remap_ok and ok
            if ok:
                self.logger.log_good('  Remap window configuration is correct: REMAPBASE <= REMAPLIMIT < TOUUD')
            else:
                self.logger.log_bad('  Remap window configuration is not correct')

        ok = (0 == tolud & ALIGNED_1MB) and \
             (0 == touud & ALIGNED_1MB) and \
             (0 == remapbase & ALIGNED_1MB) and \
             (0 == remaplimit & ALIGNED_1MB)
        remap_ok = remap_ok and ok
        if ok:
            self.logger.log_good('  All addresses are 1MB aligned')
        else:
            self.logger.log_bad('  Not all addresses are 1MB aligned')

        self.logger.log('[*] Checking if memory remap configuration is locked..')
        ok = (0 != touud_lock) or (0 != ia_untrusted)
        remap_ok = remap_ok and ok
        if ok:
            self.logger.log_good('  TOUUD is locked')
        else:
            self.logger.log_bad('  TOUUD is not locked')

        ok = (0 != tolud_lock) or (0 != ia_untrusted)
        remap_ok = remap_ok and ok
        if ok:
            self.logger.log_good('  TOLUD is locked')
        else:
            self.logger.log_bad('  TOLUD is not locked')

        if self.cs.register.has_field(remapbase_register, 'LOCK') and self.cs.register.has_field(remaplimit_register, 'LOCK'):
            ok = ((0 != remapbase_lock) and (0 != remaplimit_lock)) or (0 != ia_untrusted)
            remap_ok = remap_ok and ok
            if ok:
                self.logger.log_good('  REMAPBASE and REMAPLIMIT are locked')
            else:
                self.logger.log_bad('  REMAPBASE and REMAPLIMIT are not locked')

        if remap_ok:
            if is_warning:
                self.logger.log_warning('Most Memory Remap registers are configured correctly and locked')
                self.logger.log('[!] Manual verification of REMAP BASE and LIMIT register values may be needed.')
                res = ModuleResult.WARNING
                self.result.setStatusBit(self.result.status.VERIFY)
            else:
                res = ModuleResult.PASSED
                self.result.setStatusBit(self.result.status.SUCCESS)
                self.logger.log_passed('Memory Remap is configured correctly and locked')
        else:
            res = ModuleResult.FAILED
            self.result.setStatusBit(self.result.status.CONFIGURATION)
            self.result.setStatusBit(self.result.status.LOCKS)
            self.logger.log_failed('Memory Remap is not properly configured/locked. Remap attack may be possible')

        return self.result.getReturnCode(res)

    def run(self, _) -> int:
        self.logger.start_test('Memory Remapping Configuration')

        self.res = self.check_remap_config()
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com

from collections import namedtuple
from enum import Enum
from chipsec.library.logger import logger


class Stage(Enum):
    NONE = 0
    GET_INFO = 10
    DEVICE_CFG = 20
    CORE_SUPPORT = 30
    CUST_SUPPORT = 40
    EXTRA = 50


# Stage - None
# - Never runs
# - stage_data - None
# - Returns None

# Stage.GET_INFO
# - Gathers platform information including values used in platform detection
# - stage_data - stage_info named tuple
# - Returns - info_data named tuple
stage_info = namedtuple('StageInfo', ['vid_str', 'configuration'])
info_data = namedtuple('InfoData', ['family', 'proc_code', 'pch_code', 'detect_vals', 'req_pch', 'vid_str', 'sku_list'])

# Stage.DEVICE_CFG
# - Determine device configuration files
# - stage_data - stage_dev named tuple for file being processed
# - Returns - A list of config_data named tuples
stage_dev = namedtuple('StageCore', ['vid_str', 'xml_file'])
config_data = namedtuple('DevData', ['vid_str', 'dev_name', 'xml_file'])

# Stage.CORE_SUPPORT
# - Parse all core XML tags and update configuration data directly in object
# - stage_data - config_data named tuple for the file being processed
# - Returns - None

# Stage.CUST_SUPPORT
# - Parse any custom XML tags and update configuration data directly in object
# - stage_data - config_data named tuple for the file being processed
# - Returns - None


class BaseConfigParser:
    def __init__(self, cfg_obj):
        self.logger = logger()
        self.cfg = cfg_obj

    def startup(self):
        return None

    def get_metadata(self):
        return {'template': self.def_handler}

    def get_stage(self):
        return Stage.NONE

    def def_handler(self, et_node, stage_data=None):
        return None


parsers = [BaseConfigParser]


class BaseConfigHelper:
    def __init__(self, cfg_obj):
        self.logger = logger()
        self.cfg = cfg_obj



# -*- coding: utf-8 -*-
# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2021, SentinelOne
# Copyright (c) 2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
The Windows SMM Security Mitigation Table (WSMT) is an ACPI table defined by Microsoft that allows
system firmware to confirm to the operating system that certain security best practices have been
implemented in System Management Mode (SMM) software.

Reference:
    - See <https://docs.microsoft.com/en-us/windows-hardware/design/device-experiences/oem-uefi-wsmt> for more details.

Usage:
    ``chipsec_main -m common.wsmt``

Examples:
    >>> chipsec_main.py -m common.wsmt

.. note::
    - Analysis is only necessary if Windows is the primary OS

"""
from chipsec.module_common import BaseModule, BIOS, SMM
from chipsec.library.returncode import ModuleResult
from chipsec.hal.acpi import ACPI
from chipsec.hal.acpi_tables import WSMT

TAGS = [BIOS, SMM]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'WSMT']


class wsmt(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)
        self._acpi = ACPI(self.cs)

    def is_supported(self):
        return True

    def check_wsmt(self):
        table_data = self._acpi.get_ACPI_table('WSMT')
        if not table_data:
            self.logger.log_warning('WSMT table was not found.')
            self.result.setStatusBit(self.result.status.UNSUPPORTED_FEATURE)
            return self.result.getReturnCode(ModuleResult.WARNING)

        wsmt_table = WSMT()
        try:
            wsmt_table.parse(table_data[0][1])
        except TypeError:
            self.logger.log_error('Issue parsing the WSMT table data.')
            self.result.setStatusBit(self.result.status.PARSE_ISSUE)
            return self.result.getReturnCode(ModuleResult.ERROR)

        self.logger.log(wsmt_table)

        if (not wsmt_table.fixed_comm_buffers) or (not wsmt_table.comm_buffer_nested_ptr_protection) or (not wsmt_table.system_resource_protection):
            self.logger.log_warning('WSMT table is present but certain mitigations are missing.')
            self.result.setStatusBit(self.result.status.MITIGATION)
            self.res = self.result.getReturnCode(ModuleResult.WARNING)
        else:
            self.logger.log_passed('WSMT table is present and reports all supported mitigations.')
            self.result.setStatusBit(self.result.status.SUCCESS)
            self.res = self.result.getReturnCode(ModuleResult.PASSED)

    def run(self, module_argv):
        self.logger.start_test('WSMT Configuration')
        self.check_wsmt()
        if self.res == ModuleResult.WARNING:
            self.logger.log_important('Manual analysis of SMI handlers is required to determine if they can be abused by attackers to circumvent VBS.')
        return self.res



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
import os
import tempfile
import unittest

from tests.software import mock_helper

import chipsec_util
from chipsec import chipset


class TestChipsecUtil(unittest.TestCase):
    """Test the commands exposed by chipsec_utils.

    Each test may define its virtual helper and then call the _chipsec_util
    method with the command line arguments.
    """

    def setUp(self):
        """Setup the environment for the utils tests.

        We mock the helper registry to only contain our emulated helper.
        """
        fileno, self.log_file = tempfile.mkstemp()
        os.close(fileno)
        chipset._chipset = None

    def tearDown(self):
        os.remove(self.log_file)
        chipset._chipset = None
        chipsec_util._cs = None

    def _chipsec_util(self, arg, helper_class=mock_helper.TestHelper):
        """Run the chipsec_util command with the arguments.

        Each test may setup a virtual helper to emulate the expected behaviour
        from the hardware. If no helper is provided, TestHelper will be used.
        It verifies that no error is being reported. self.log will be populated
        with the output.
        """
        args = arg.split()
        par = chipsec_util.parse_args(args)
        util = chipsec_util.ChipsecUtil(par, args)
        util._helper = helper_class()
        util.logger.VERBOSE = True
        util.logger.HAL = True
        util.logger.set_log_file(self.log_file)
        try:
            err_code = util.main()
        finally:
            util.logger.close()
        with open(self.log_file, 'rb') as log:
            self.log = log.read()
        self.assertEqual(err_code, 0)

    def _assertLogValue(self, name, value):
        """Shortcut to validate the output.

        Assert that at least one line exists within the log which matches the
        expression: name [:=] value.
        """
        exp = r'(^|\W){}\s*[:=]\s*{}($|\W)'.format(name, value)
        self.assertRegex(self.log, exp.encode())



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2024, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

from chipsec.library.logger import logger


class Control:

    def __init__(self, cs) -> None:
        self.cs = cs

    def get(self, control_name: str, cpu_thread: int = 0, with_print: bool = False) -> int:
        """Reads some control field (by name)"""
        control = self.cs.Cfg.CONTROLS[control_name]
        reg = control['register']
        field = control['field']
        reg_data = self.cs.register.read(reg, cpu_thread)
        if logger().VERBOSE or with_print:
            self.cs.register.print(reg, reg_data)
        return self.cs.register.get_field(reg, reg_data, field)

    def set(self, control_name: str, control_value: int, cpu_thread: int = 0) -> bool:
        """Writes some control field (by name)"""
        control = self.cs.Cfg.CONTROLS[control_name]
        reg = control['register']
        field = control['field']
        return self.cs.register.write_field(reg, field, control_value, cpu_thread=cpu_thread)

    def is_defined(self, control_name: str) -> bool:
        """Returns True if control_name Control is defined."""
        try:
            return (self.cs.Cfg.CONTROLS[control_name] is not None)
        except KeyError:
            return False

    def is_all_ffs(self, control_name: str, cpu_thread: int = 0, field_only: bool = False) -> bool:
        """Returns True if control_name Control value is all 0xFFs (all 1's)"""
        if self.is_defined(control_name) is None:
            if logger().DEBUG:
                logger().log_error(f"Control '{control_name}' not defined.")
            return True
        control = self.cs.Cfg.CONTROLS[control_name]
        reg_def = control['register']
        reg_data = self.cs.register.read(reg_def, cpu_thread)
        if field_only:
            reg_field = control['field']
            reg_data = self.cs.register.get_field(reg_def, reg_data, reg_field)
            result = self.cs.register.is_field_all_ones(reg_def, reg_field, reg_data)
        else:
            result = self.cs.register.is_all_ffs(reg_def, reg_data)
        return result



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

""""
To execute: python[3] -m unittest tests.utilcmd.acpi_cmd.test_acpi_cmd
"""

import os
import struct
import unittest

from chipsec.library.file import get_main_dir
from tests.utilcmd.run_chipsec_util import setup_run_destroy_util
from chipsec.testcase import ExitCode
from tests.software import mock_helper, util


class TestAcpiUtilcmd(unittest.TestCase):
    def test_list(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        acpi_list_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "acpi_cmd", "acpi_cmd_list_1.json")
        retval = setup_run_destroy_util(init_replay_file, "acpi", "list", util_replay_file=acpi_list_replay_file)
        self.assertEqual(retval, ExitCode.OK)

    def test_table(self) -> None:
        init_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "adlenumerate.json")
        acpi_table_replay_file = os.path.join(get_main_dir(), "tests", "utilcmd", "acpi_cmd", "acpi_cmd_table_1.json")
        retval = setup_run_destroy_util(init_replay_file, "acpi", "table XSDT", util_replay_file=acpi_table_replay_file)
        self.assertEqual(retval, ExitCode.OK)


class TestACPIChipsecUtil(util.TestChipsecUtil):
    def test_acpi_xsdt_list(self):

        class ACPIHelper(mock_helper.ACPIHelper):

            USE_RSDP_REV_0 = False

            def __init__(self):
                super(ACPIHelper, self).__init__()
                self._add_entry_to_xsdt(0x400)

            def read_phys_mem(self, pa, length):
                if (pa & 0xffffffff) == 0x400:
                    return b"EFGH"
                else:
                    parent = super(ACPIHelper, self)
                    return parent.read_phys_mem(pa, length)

        self._chipsec_util("acpi list", ACPIHelper)
        self._assertLogValue("EFGH", "0x0000000000000400")

    def test_acpi_rsdt_list(self):

        class ACPIHelper(mock_helper.ACPIHelper):

            def __init__(self):
                super(ACPIHelper, self).__init__()
                self._add_entry_to_rsdt(0x300)

            def read_phys_mem(self, pa, length):
                pa_lo = pa & 0xFFFFFFFF
                if (pa_lo >= self.EBDA_ADDRESS and
                        pa_lo < self.RSDP_ADDRESS + len(self.rsdp_descriptor)):
                    # Simulate a condition where there is no RSDP in EBDA
                    return b"\xFF" * length
                elif pa_lo == 0xE0000:
                    return self.rsdp_descriptor[:length]
                elif pa_lo == 0x300:
                    return b"ABCD"
                else:
                    parent = super(ACPIHelper, self)
                    return parent.read_phys_mem(pa, length)

        self._chipsec_util("acpi list", ACPIHelper)
        self._assertLogValue("ABCD", "0x0000000000000300")

    def test_acpi_facp_list(self):
        self._chipsec_util("acpi table FACP", mock_helper.DSDTParsingHelper)
        self._assertLogValue("DSDT", "0x00000000")
        self._assertLogValue("X_DSDT", "0x0000000000000000")

    def test_mismatch_dsdt_x_dsdt_error(self):

        class DSDTParsingHelper(mock_helper.DSDTParsingHelper):

            DSDT_ADDRESS = 0x400
            X_DSDT_ADDRESS = 0x312

        self._chipsec_util("acpi table FACP", DSDTParsingHelper)
        self._assertLogValue("DSDT", "0x00000400")
        self._assertLogValue("X_DSDT", "0x0000000000000312")
        self.assertIn(b"Unable to determine the correct DSDT address", self.log)

    def test_mismatch_dsdt_x_dsdt_ok_dsdt_zero(self):

        class DSDTParsingHelper(mock_helper.DSDTParsingHelper):

            DSDT_ADDRESS = 0x0
            X_DSDT_ADDRESS = 0x400

        self._chipsec_util("acpi table FACP", DSDTParsingHelper)
        self._assertLogValue("DSDT", "0x00000000")
        self._assertLogValue("X_DSDT", "0x0000000000000400")
        self.assertNotIn(b"Unable to determine the correct DSDT address",
                         self.log)

    def test_mismatch_dsdt_x_dsdt_ok_x_dsdt_zero(self):

        class DSDTParsingHelper(mock_helper.DSDTParsingHelper):

            DSDT_ADDRESS = 0x400
            X_DSDT_ADDRESS = 0x0

        self._chipsec_util("acpi table FACP", DSDTParsingHelper)
        self._assertLogValue("DSDT", "0x00000400")
        self._assertLogValue("X_DSDT", "0x0000000000000000")
        self.assertNotIn(b"Unable to determine the correct DSDT address",
                         self.log)

    def test_no_x_dsdt(self):

        class DSDTParsingHelper(mock_helper.DSDTParsingHelper):

            USE_FADT_WITH_X_DSDT = False
            DSDT_ADDRESS = 0x400

        self._chipsec_util("acpi table FACP", DSDTParsingHelper)
        self._assertLogValue("DSDT", "0x00000400")
        self._assertLogValue("X_DSDT", "Not found")
        self.assertIn(b"Cannot find X_DSDT entry in FADT.", self.log)

    def test_show_dsdt(self):

        class DSDTParsingHelper(mock_helper.DSDTParsingHelper):

            DSDT_ADDRESS = 0x600

            DSDT_DESCRIPTOR = (b"DSDT" +                   # Signature
                               struct.pack("<I", 0x30) +  # Length
                               struct.pack("<B", 0x1) +   # Revision
                               struct.pack("<B", 0x1) +   # Checksum
                               b"OEMDSD" +                 # OEMID
                               b"OEMTBLID" +               # OEM Table ID
                               b"OEMR" +                   # OEM Revision
                               b"CRID" +                   # Creator ID
                               b"CRRV" +                   # Creator Revision
                               struct.pack("<Q", 0x129))  # AML code

            def read_phys_mem(self, pa, length):
                pa_lo = pa & 0xFFFFFFFF
                if pa_lo == self.DSDT_ADDRESS:
                    return self.DSDT_DESCRIPTOR[:length]
                else:
                    parent = super(DSDTParsingHelper, self)
                    return parent.read_phys_mem(pa, length)

        self._chipsec_util("acpi table DSDT", DSDTParsingHelper)
        self.assertIn(b"OEMDSD", self.log)

    def test_parse_multi_table(self):
        """Test to verify that tables with same signature are parsed correctly.

        Since usually there are several SSDT tables with the
        same signature, we test SSDT parsing.
        """

        class SSDTParsingHelper(mock_helper.ACPIHelper):
            """Test helper containing generic descriptor for SSDT to parse SSDT

            Three regions are defined:
              * SSDT table [0x400, 0x430]
              * SSDT table [0x600, 0x630]
              * SSDT table [0x800, 0x830]
            """
            SSDT1_DESCRIPTOR = (b"SSDT" +                   # Signature
                                struct.pack("<I", 0x30) +  # Length
                                struct.pack("<B", 0x1) +   # Revision
                                struct.pack("<B", 0x1) +   # Checksum
                                b"OEMSS1" +                 # OEMID
                                b"OEMTBLID" +               # OEM Table ID
                                b"OEMR" +                   # OEM Revision
                                b"CRID" +                   # Creator ID
                                b"CRRV" +                   # Creator Revision
                                struct.pack("<Q", 0x129))  # AML code

            SSDT2_DESCRIPTOR = (b"SSDT" +                   # Signature
                                struct.pack("<I", 0x30) +  # Length
                                struct.pack("<B", 0x1) +   # Revision
                                struct.pack("<B", 0x1) +   # Checksum
                                b"OEMSS2" +                 # OEMID
                                b"OEMTBLID" +               # OEM Table ID
                                b"OEMR" +                   # OEM Revision
                                b"CRID" +                   # Creator ID
                                b"CRRV" +                   # Creator Revision
                                struct.pack("<Q", 0x929))  # AML code

            SSDT3_DESCRIPTOR = (b"SSDT" +                   # Signature
                                struct.pack("<I", 0x30) +  # Length
                                struct.pack("<B", 0x1) +   # Revision
                                struct.pack("<B", 0x1) +   # Checksum
                                b"OEMSS3" +                 # OEMID
                                b"OEMTBLID" +               # OEM Table ID
                                b"OEMR" +                   # OEM Revision
                                b"CRID" +                   # Creator ID
                                b"CRRV" +                   # Creator Revision
                                struct.pack("<Q", 0x199))  # AML code

            def __init__(self):
                super(SSDTParsingHelper, self).__init__()
                self._add_entry_to_rsdt(0x400)
                self._add_entry_to_rsdt(0x600)
                self._add_entry_to_rsdt(0x800)

            def read_phys_mem(self, pa, length):
                pa_lo = pa & 0xFFFFFFFF
                if pa_lo == 0x400:
                    return self.SSDT1_DESCRIPTOR[:length]
                elif pa_lo == 0x600:
                    return self.SSDT2_DESCRIPTOR[:length]
                elif pa_lo == 0x800:
                    return self.SSDT3_DESCRIPTOR[:length]
                else:
                    parent = super(SSDTParsingHelper, self)
                    return parent.read_phys_mem(pa, length)

        self._chipsec_util("acpi list", SSDTParsingHelper)
        self._assertLogValue("SSDT", ("0x0000000000000400, "
                                      "0x0000000000000600, "
                                      "0x0000000000000800"))
        self._chipsec_util("acpi table SSDT", SSDTParsingHelper)
        self.assertIn(b"OEMSS1", self.log)
        self.assertIn(b"OEMSS2", self.log)
        self.assertIn(b"OEMSS3", self.log)


if __name__ == '__main__':
    unittest.main()
